import mock
from click.testing import CliRunner

from gradient.api_sdk import constants
from gradient.api_sdk.clients.http_client import default_headers
from gradient.cli import cli
from tests import MockResponse, example_responses

EXPECTED_HEADERS = default_headers.copy()
EXPECTED_HEADERS["ps_client_name"] = "gradient-cli"

EXPECTED_HEADERS_WITH_CHANGED_API_KEY = EXPECTED_HEADERS.copy()
EXPECTED_HEADERS_WITH_CHANGED_API_KEY["X-API-Key"] = "some_key"


class TestCreateHyperparameters(object):
    URL_V2 = "https://services.paperspace.io/experiments/v2/hyperopt/"
    TAGS_URL = "https://api.paperspace.io/entityTags/updateTags"
    COMMAND = [
        "experiments", "hyperparameters", "create",
        "--name", "some_name",
        "--tuningCommand", "some command",
        "--workerContainer", "some_container",
        "--workerMachineType", "k80",
        "--workerCommand", "some worker command",
        "--workerCount", "1",
        "--projectId", "some_project_id",
    ]
    COMMAND_WITH_TAGS = [
        "experiments", "hyperparameters", "create",
        "--name", "some_name",
        "--tuningCommand", "some command",
        "--workerContainer", "some_container",
        "--workerMachineType", "k80",
        "--workerCommand", "some worker command",
        "--workerCount", "1",
        "--projectId", "some_project_id",
        "--tag", "test0",
        "--tag", "test1",
        "--tags", "test2,test3",
    ]
    EXPECTED_REQUEST_JSON = {
        "workerContainer": "some_container",
        "workerMachineType": "k80",
        "name": "some_name",
        "tuningCommand": "c29tZSBjb21tYW5k",
        "workerCount": 1,
        "workerCommand": "c29tZSB3b3JrZXIgY29tbWFuZA==",
        "experimentTypeId": constants.ExperimentType.HYPERPARAMETER_TUNING,
        "projectHandle": "some_project_id",
    }
    TAGS_JSON = {
        "entity": "experiment",
        "entityId": "eshgvasywz9k1w",
        "tags": ["test0", "test1", "test2", "test3"]
    }
    UPDATE_TAGS_RESPONSE_JSON_200 = example_responses.UPDATE_TAGS_RESPONSE

    COMMAND_WHEN_ALL_PARAMETERS_WERE_USED = [
        "experiments", "hyperparameters", "create",
        "--name", "some_name",
        "--tuningCommand", "some command",
        "--workerContainer", "some_worker_container",
        "--workerMachineType", "k80",
        "--workerCommand", "some worker command",
        "--workerCount", "666",
        "--projectId", "some_project_id",
        "--workerRegistryUsername", "some_registry_username",
        "--workerRegistryPassword", "some_registry_password",
        "--workerContainerUser", "some_worker_container_user",
        "--hyperparameterServerRegistryUsername", "some_hyperparameter_registry_username",
        "--hyperparameterServerRegistryPassword", "some_hyperparameter_registry_password",
        "--hyperparameterServerContainer", "some_hyperparameter_container",
        "--hyperparameterServerContainerUser", "some_hyperparameter_container_user",
        "--hyperparameterServerMachineType", "some_hyperparameter_server_machine",
        "--modelPath", "some_model_path",
        "--modelType", "some_model_type",
        "--ignoreFiles", "file1,file2",
        "--isPreemptible",
        "--artifactDirectory", "some_artifact_directory",
        "--clusterId", "some_cluster_id",
        "--experimentEnv", "{\"key\":\"val\"}",
        "--ignoreFiles", "file2",
        "--ports", "8080,9000:9999",
        "--workerDockerfilePath", "some_docker_path",
        "--workerUseDockerfile",
        "--workingDirectory", "some_working_directory",
        "--workspace", "s3://some-path",
    ]
    EXPECTED_REQUEST_JSON_WHEN_ALL_PARAMETERS_WERE_USED = {
        "workerContainer": "some_worker_container",
        "workerMachineType": "k80",
        "name": "some_name",
        "tuningCommand": "c29tZSBjb21tYW5k",
        "workerCount": 666,
        "workerCommand": "c29tZSB3b3JrZXIgY29tbWFuZA==",
        "workerRegistryUsername": "some_registry_username",
        "workerRegistryPassword": "some_registry_password",
        "workerContainerUser": "some_worker_container_user",
        "projectHandle": "some_project_id",
        "hyperparameterServerRegistryUsername": "some_hyperparameter_registry_username",
        "hyperparameterServerRegistryPassword": "some_hyperparameter_registry_password",
        "hyperparameterServerContainer": "some_hyperparameter_container",
        "hyperparameterServerContainerUser": "some_hyperparameter_container_user",
        "hyperparameterServerMachineType": "some_hyperparameter_server_machine",
        "experimentTypeId": constants.ExperimentType.HYPERPARAMETER_TUNING,
        "modelPath": "some_model_path",
        "modelType": "some_model_type",
        "isPreemptible": True,
        "dockerfilePath": "some_docker_path",
        "artifactDirectory": "some_artifact_directory",
        "clusterId": "some_cluster_id",
        "experimentEnv": {"key": "val"},
        "ports": "8080,9000:9999",
        "useDockerfile": True,
        "workingDirectory": "some_working_directory",
        "workspaceUrl": "s3://some-path",
    }
    COMMAND_WITH_OPTIONS_FILE = ["experiments", "hyperparameters", "create", "--optionsFile", ]  # path added in test

    EXPECTED_RESPONSE = {"handle": "eshgvasywz9k1w", "message": "success"}
    EXPECTED_STDOUT = "Hyperparameter tuning job created with ID: eshgvasywz9k1w\n" \
                      "https://console.paperspace.com/projects/some_project_id/experiments/eshgvasywz9k1w\n"

    EXPECTED_RESPONSE_JSON_WITH_ERROR = {
        "details": {
            "projectHandle": ["Missing data for required field."],
        },
        "error": "Experiment data error",
    }
    EXPECTED_STDOUT_WHEN_ERROR_RECEIVED = "Failed to create resource: projectHandle: Missing data for required field." \
                                          "\nExperiment data error\n"

    COMMAND_WITH_API_KEY_PARAMETER_USED = [
        "experiments", "hyperparameters", "create",
        "--name", "some_name",
        "--tuningCommand", "some command",
        "--workerContainer", "some_container",
        "--workerMachineType", "k80",
        "--workerCommand", "some worker command",
        "--workerCount", "1",
        "--projectId", "some_project_id",
        "--apiKey", "some_key",
    ]
    EXPECTED_REQUEST_JSON_WHEN_API_KEY_PARAMETERS_WAS_USED = {
        "workerContainer": "some_container",
        "workerMachineType": "k80",
        "name": "some_name",
        "tuningCommand": "some command",
        "workerCount": 1,
        "workerCommand": "some worker command",
        "projectHandle": "pr4yxj956",
        "experimentTypeId": constants.ExperimentType.HYPERPARAMETER_TUNING,
    }

    EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED = {"details": "Incorrect API Key provided", "error": "Forbidden"}
    EXPECTED_STDOUT_WHEN_WRONG_API_KEY_WAS_USED = "Failed to create resource: Incorrect API Key provided\nForbidden\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_get_request_and_print_proper_message_when_create_command_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE, 201)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_get_request_and_print_proper_message_when_create_command_was_used_with_all_options(self,
                                                                                                            post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE, 201)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WHEN_ALL_PARAMETERS_WERE_USED)

        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON_WHEN_ALL_PARAMETERS_WERE_USED,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_replace_api_key_in_headers_when_api_key_parameter_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE, 201)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_PARAMETER_USED)

        assert result.output == self.EXPECTED_STDOUT
        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_proper_message_when_error_message_received(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON_WITH_ERROR, 400)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == self.EXPECTED_STDOUT_WHEN_ERROR_RECEIVED

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_proper_message_when_wrong_api_key_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED, 403)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == self.EXPECTED_STDOUT_WHEN_WRONG_API_KEY_WAS_USED
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_yaml_file(self, post_patched, hyperparameters_create_config_path):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [hyperparameters_create_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=self.EXPECTED_REQUEST_JSON_WHEN_ALL_PARAMETERS_WERE_USED,
                                             params=None,
                                             files=None,
                                             data=None)

        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_request_and_print_proper_message_when_error_code_returned_without_json_data(self,
                                                                                                     post_patched):
        post_patched.return_value = MockResponse(status_code=500)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == "Failed to create resource\n"
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_proper_data_and_tag_hyperopt_experiment(self, post_patched, get_patched, put_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE)
        get_patched.return_value = MockResponse({}, )
        put_patched.return_value = MockResponse(self.UPDATE_TAGS_RESPONSE_JSON_200)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_TAGS)

        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        put_patched.assert_called_once_with(
            self.TAGS_URL,
            headers=EXPECTED_HEADERS,
            json=self.TAGS_JSON,
            params=None,
            data=None,
        )

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        assert result.exit_code == 0


class TestCreateAndStartHyperparameters(object):
    URL_V2 = "https://services.paperspace.io/experiments/v2/hyperopt/create_and_start/"
    COMMAND = [
        "experiments", "hyperparameters", "run",
        "--name", "some_name",
        "--tuningCommand", "some command",
        "--workerContainer", "some_container",
        "--workerMachineType", "k80",
        "--workerCommand", "some worker command",
        "--workerCount", "1",
        "--projectId", "some_project_id",
    ]
    EXPECTED_REQUEST_JSON = {
        "workerContainer": "some_container",
        "workerMachineType": "k80",
        "name": "some_name",
        "tuningCommand": "c29tZSBjb21tYW5k",
        "workerCount": 1,
        "workerCommand": "c29tZSB3b3JrZXIgY29tbWFuZA==",
        "projectHandle": "some_project_id",
        "experimentTypeId": constants.ExperimentType.HYPERPARAMETER_TUNING,
    }

    COMMAND_WHEN_ALL_PARAMETERS_WERE_USED = [
        "experiments", "hyperparameters", "run",
        "--name", "some_name",
        "--tuningCommand", "some command",
        "--workerContainer", "some_worker_container",
        "--workerMachineType", "k80",
        "--workerCommand", "some worker command",
        "--workerCount", "666",
        "--projectId", "some_project_id",
        "--workerRegistryUsername", "some_registry_username",
        "--workerRegistryPassword", "some_registry_password",
        "--workerContainerUser", "some_worker_container_user",
        "--hyperparameterServerRegistryUsername", "some_hyperparameter_registry_username",
        "--hyperparameterServerRegistryPassword", "some_hyperparameter_registry_password",
        "--hyperparameterServerContainer", "some_hyperparameter_container",
        "--hyperparameterServerContainerUser", "some_hyperparameter_container_user",
        "--hyperparameterServerMachineType", "some_hyperparameter_server_machine",
        "--modelPath", "some_model_path",
        "--modelType", "some_model_type",
        "--ignoreFiles", "file1,file2",
        "--isPreemptible",
        "--artifactDirectory", "some_artifact_directory",
        "--clusterId", "some_cluster_id",
        "--experimentEnv", "{\"key\":\"val\"}",
        "--ignoreFiles", "file2",
        "--ports", "8080,9000:9999",
        "--workerDockerfilePath", "some_docker_path",
        "--workerUseDockerfile",
        "--workingDirectory", "some_working_directory",
        "--workspace", "s3://some-path",
    ]
    EXPECTED_REQUEST_JSON_WHEN_ALL_PARAMETERS_WERE_USED = {
        "workerContainer": "some_worker_container",
        "workerMachineType": "k80",
        "name": "some_name",
        "tuningCommand": "c29tZSBjb21tYW5k",
        "workerCount": 666,
        "workerCommand": "c29tZSB3b3JrZXIgY29tbWFuZA==",
        "workerRegistryUsername": "some_registry_username",
        "workerRegistryPassword": "some_registry_password",
        "workerContainerUser": "some_worker_container_user",
        "projectHandle": "some_project_id",
        "hyperparameterServerRegistryUsername": "some_hyperparameter_registry_username",
        "hyperparameterServerRegistryPassword": "some_hyperparameter_registry_password",
        "hyperparameterServerContainer": "some_hyperparameter_container",
        "hyperparameterServerContainerUser": "some_hyperparameter_container_user",
        "hyperparameterServerMachineType": "some_hyperparameter_server_machine",
        "experimentTypeId": constants.ExperimentType.HYPERPARAMETER_TUNING,
        "modelPath": "some_model_path",
        "modelType": "some_model_type",
        "isPreemptible": True,
        "dockerfilePath": "some_docker_path",
        "artifactDirectory": "some_artifact_directory",
        "clusterId": "some_cluster_id",
        "experimentEnv": {"key": "val"},
        "ports": "8080,9000:9999",
        "useDockerfile": True,
        "workingDirectory": "some_working_directory",
        "workspaceUrl": "s3://some-path",
    }
    COMMAND_WITH_OPTIONS_FILE = ["experiments", "hyperparameters", "run", "--optionsFile", ]  # path added in test

    EXPECTED_RESPONSE = {"handle": "eshgvasywz9k1w", "message": "success"}
    EXPECTED_STDOUT = "Hyperparameter tuning job created and started with ID: eshgvasywz9k1w\n" \
                      "https://console.paperspace.com/projects/some_project_id/experiments/eshgvasywz9k1w\n"

    EXPECTED_RESPONSE_JSON_WITH_ERROR = {
        "details": {
            "projectHandle": ["Missing data for required field."],
        },
        "error": "Experiment data error",
    }
    EXPECTED_STDOUT_WHEN_ERROR_RECEIVED = "Failed to create resource: " \
                                          "projectHandle: Missing data for required field.\nExperiment data error\n"

    COMMAND_WITH_API_KEY_PARAMETER_USED = [
        "experiments", "hyperparameters", "run",
        "--name", "some_name",
        "--tuningCommand", "some command",
        "--workerContainer", "some_container",
        "--workerMachineType", "k80",
        "--workerCommand", "some worker command",
        "--workerCount", "1",
        "--projectId", "some_project_id",
        "--apiKey", "some_key",
    ]
    EXPECTED_REQUEST_JSON_WHEN_API_KEY_PARAMETERS_WAS_USED = {
        "workerContainer": "some_container",
        "workerMachineType": "k80",
        "name": "some_name",
        "tuningCommand": "some command",
        "workerCount": 1,
        "workerCommand": "some worker command",
        "projectHandle": "pr4yxj956",
        "experimentTypeId": constants.ExperimentType.HYPERPARAMETER_TUNING,
    }

    EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED = {"details": "Incorrect API Key provided", "error": "Forbidden"}
    EXPECTED_STDOUT_WHEN_WRONG_API_KEY_WAS_USED = "Failed to create resource: Incorrect API Key provided\nForbidden\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_get_request_and_print_proper_message_when_create_command_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE, 201)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_get_request_and_print_proper_message_when_create_command_was_used_with_all_options(self,
                                                                                                            post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE, 201)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WHEN_ALL_PARAMETERS_WERE_USED)

        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON_WHEN_ALL_PARAMETERS_WERE_USED,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_replace_api_key_in_headers_when_api_key_parameter_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE, 201)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_PARAMETER_USED)

        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_proper_message_when_error_message_received(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON_WITH_ERROR, 400)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == self.EXPECTED_STDOUT_WHEN_ERROR_RECEIVED

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_proper_message_when_wrong_api_key_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED, 403)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == self.EXPECTED_STDOUT_WHEN_WRONG_API_KEY_WAS_USED
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_request_and_print_proper_message_when_error_code_returned_without_json_data(self,
                                                                                                     post_patched):
        post_patched.return_value = MockResponse(status_code=500)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == "Failed to create resource\n"
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_yaml_file(self, post_patched, hyperparameters_create_config_path):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [hyperparameters_create_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=self.EXPECTED_REQUEST_JSON_WHEN_ALL_PARAMETERS_WERE_USED,
                                             params=None,
                                             files=None,
                                             data=None)
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"


class TestStartHyperparameters(object):
    URL_V2 = "https://services.paperspace.io/experiments/v2/hyperopt/some_id/start/"
    COMMAND = [
        "experiments", "hyperparameters", "start",
        "--id", "some_id",
    ]

    COMMAND_WITH_OPTIONS_FILE = ["experiments", "hyperparameters", "start", "--optionsFile", ]  # path added in test

    EXPECTED_RESPONSE = {"message": "success"}
    EXPECTED_STDOUT = "Hyperparameter tuning started\n"

    EXPECTED_RESPONSE_JSON_WITH_ERROR = {"error": "Could not find cluster meeting requirements"}
    EXPECTED_STDOUT_WHEN_ERROR_RECEIVED = "Failed to start hyperparameter tuning job: Could not find cluster meeting requirements\n"

    COMMAND_WITH_API_KEY_PARAMETER_USED = [
        "experiments", "hyperparameters", "start",
        "--id", "some_id",
        "--apiKey", "some_key",
    ]
    EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED = {"details": "Incorrect API Key provided", "error": "Forbidden"}
    EXPECTED_STDOUT_WHEN_WRONG_API_KEY_WAS_USED = "Failed to start hyperparameter tuning job: Incorrect API Key provided\nForbidden\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    def test_should_send_get_request_and_print_proper_message_when_start_command_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE, 201)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS,
                                             json=None,
                                             params=None,
                                             data=None,
                                             )
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    def test_should_replace_api_key_in_headers_when_api_key_parameter_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE, 201)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_PARAMETER_USED)

        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=None,
                                             params=None,
                                             data=None,
                                             )

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    def test_should_read_options_from_yaml_file(self, post_patched, hyperparameters_start_config_path):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE, 201)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [hyperparameters_start_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL_V2,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=None,
                                             params=None,
                                             data=None,
                                             )

        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    def test_should_print_proper_message_when_error_message_received(self, put_patched):
        put_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON_WITH_ERROR, 400)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        put_patched.assert_called_once_with(self.URL_V2,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=None,
                                            data=None,
                                            )

        assert result.output == self.EXPECTED_STDOUT_WHEN_ERROR_RECEIVED

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    def test_should_print_proper_message_when_wrong_api_key_was_used(self, put_patched):
        put_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED, 403)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        put_patched.assert_called_once_with(self.URL_V2,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=None,
                                            data=None,
                                            )

        assert result.output == self.EXPECTED_STDOUT_WHEN_WRONG_API_KEY_WAS_USED
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    def test_should_send_request_and_print_proper_message_when_error_code_returned_without_json_data(self, put_patched):
        put_patched.return_value = MockResponse(status_code=500)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        put_patched.assert_called_once_with(self.URL_V2,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=None,
                                            data=None,
                                            )

        assert result.output == "Failed to start hyperparameter tuning job\n"
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"


class TestHyperparametersList(object):
    URL_V2 = "https://services.paperspace.io/experiments/v2/hyperopt/"
    COMMAND = ["experiments", "hyperparameters", "list"]
    COMMAND_WITH_OPTIONS_FILE = ["experiments", "hyperparameters", "list", "--optionsFile", ]  # path added in test
    EXPECTED_REQUEST_PARAMS = {"limit": -1}

    COMMAND_WITH_API_KEY_PARAMETER_USED = ["experiments", "hyperparameters", "list", "--apiKey", "some_key"]

    EXPECTED_RESPONSE_JSON_WHEN_NO_OBJECTS_WERE_FOUND = {
        "data": [],
        "message": "success",
        "meta": {
            "filter": [],
            "limit": -1,
            "offset": 0,
            "totalItems": 0,
        },
    }

    EXPECTED_STDOUT = """+-----------+----------------+------------+
| Name      | ID             | Project ID |
+-----------+----------------+------------+
| some_name | es3dn6fu16r4kk | pr4yxj956  |
| some_name | eshlqek7wzvrxa | pr4yxj956  |
| some_name | esdwnui5qsk8qm | pr4yxj956  |
| some_name | eshz1z9k37w4nm | pr4yxj956  |
+-----------+----------------+------------+
"""

    EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED = {"status": 401, "message": "No such API token"}

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_list_of_hyperparameters(self, get_patched):
        get_patched.return_value = MockResponse(example_responses.LIST_HYPERPARAMETERS_RESPONSE_JSON, 200)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_once_with(self.URL_V2,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=self.EXPECTED_REQUEST_PARAMS)
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_replace_api_key_in_headers_when_api_key_parameter_was_used(self, get_patched):
        get_patched.return_value = MockResponse(example_responses.LIST_HYPERPARAMETERS_RESPONSE_JSON, 200)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_PARAMETER_USED)

        get_patched.assert_called_once_with(self.URL_V2,
                                            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                            json=None,
                                            params=self.EXPECTED_REQUEST_PARAMS)

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml(self, get_patched, hyperparameters_list_config_path):
        get_patched.return_value = MockResponse(example_responses.LIST_HYPERPARAMETERS_RESPONSE_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [hyperparameters_list_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        get_patched.assert_called_once_with(self.URL_V2,
                                            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                            json=None,
                                            params=self.EXPECTED_REQUEST_PARAMS)

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_proper_message_when_no_objects_were_found(
            self, get_patched):
        get_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON_WHEN_NO_OBJECTS_WERE_FOUND, 200)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_called_once_with(self.URL_V2,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=self.EXPECTED_REQUEST_PARAMS)

        assert result.output == "No data found\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_proper_message_when_wrong_api_key_was_used(self, get_patched):
        get_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED, 401)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_called_once_with(self.URL_V2,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=self.EXPECTED_REQUEST_PARAMS)

        assert result.output == "Failed to fetch data: No such API token\n"


class TestHyperparametersDetails(object):
    URL_V2 = "https://services.paperspace.io/experiments/v2/hyperopt/some_id/"
    COMMAND = ["experiments", "hyperparameters", "details", "--id", "some_id"]

    COMMAND_WITH_OPTIONS_FILE = ["experiments", "hyperparameters", "details", "--optionsFile", ]  # path added in test
    COMMAND_WITH_API_KEY_PARAMETER_USED = [
        "experiments", "hyperparameters", "details",
        "--id", "some_id",
        "--apiKey", "some_key",
    ]

    EXPECTED_RESPONSE_JSON_WHEN_NO_OBJECT_WAS_NOT_FOUND = {"error": "Hyperopt not found"}

    EXPECTED_STDOUT = """+-----------------------+---------------------+
| ID                    | ess6t3fjs2hb1g      |
+-----------------------+---------------------+
| Name                  | some_name           |
| Ports                 | 5000                |
| Project ID            | pr4yxj956           |
| Tuning command        | some command        |
| Worker command        | some worker command |
| Worker container      | some_container      |
| Worker count          | 1                   |
| Worker machine type   | k80                 |
| Worker use dockerfile | False               |
| Workspace URL         | none                |
+-----------------------+---------------------+
"""

    EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED = {"status": 401, "message": "No such API token"}

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_details_of_hyperparameters_job(self, get_patched):
        get_patched.return_value = MockResponse(example_responses.HYPERPARAMETERS_DETAILS_RESPONSE_JSON, 200)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_once_with(self.URL_V2,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=None)

        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_replace_api_key_in_headers_when_api_key_parameter_was_used(self, get_patched):
        get_patched.return_value = MockResponse(example_responses.HYPERPARAMETERS_DETAILS_RESPONSE_JSON, 200)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_PARAMETER_USED)

        get_patched.assert_called_once_with(self.URL_V2,
                                            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                            json=None,
                                            params=None)

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(
            self, get_patched, hyperparameters_details_config_path):
        get_patched.return_value = MockResponse(example_responses.HYPERPARAMETERS_DETAILS_RESPONSE_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [hyperparameters_details_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_once_with(self.URL_V2,
                                            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                            json=None,
                                            params=None)
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_proper_message_when_no_objects_were_found(self, get_patched):
        get_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON_WHEN_NO_OBJECT_WAS_NOT_FOUND, 404)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_called_once_with(self.URL_V2,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=None)

        assert result.output == "Failed to fetch data: Hyperopt not found\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_proper_message_when_wrong_api_key_was_used(self, get_patched):
        get_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED, 401)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_called_once_with(self.URL_V2,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=None)

        assert result.output == "Failed to fetch data: No such API token\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_proper_message_when_got_error_response_without_data(self, get_patched):
        get_patched.return_value = MockResponse(status_code=500)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_called_once_with(self.URL_V2,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=None)

        assert result.output == "Failed to fetch data\n"



import marshmallow

from .base import BaseSchema
from .. import models


class Model(BaseSchema):
    MODEL = models.Model

    id = marshmallow.fields.Str()
    name = marshmallow.fields.Str()
    project_id = marshmallow.fields.Str(dump_to="projectId", load_from="projectId")
    experiment_id = marshmallow.fields.Str(dump_to="experimentId", load_from="experimentId")
    tags = marshmallow.fields.List(marshmallow.fields.Str(), load_only=True)
    model_type = marshmallow.fields.Str(dump_to="modelType", load_from="modelType")
    url = marshmallow.fields.Str()
    model_path = marshmallow.fields.Str(dump_to="modelPath", load_from="modelPath")
    deployment_state = marshmallow.fields.Str(dump_to="deploymentState", load_from="deploymentState")
    summary = marshmallow.fields.Dict()
    detail = marshmallow.fields.Dict()
    notes = marshmallow.fields.Str()
    dataset_ref = marshmallow.fields.Str(dump_to="datasetRef", load_from="datasetRef")


class ModelFileSchema(BaseSchema):
    MODEL = models.ModelFile

    file = marshmallow.fields.Str()
    url = marshmallow.fields.Str()
    size = marshmallow.fields.Int()



import pytest

from gradient.api_sdk.clients.base_client import BaseClient, TagsSupportMixin


class ClassWithTagsSupportMixin(TagsSupportMixin, BaseClient):
    pass


class TestTagsSupportMixinMethods(object):
    example_entity_tags = [{"some_id": ["test0", "test2", "test1", "test3"]}]

    @pytest.mark.parametrize(
        "entity_id, entity_tags, new_tags, expected_result_tags",
        [
            ("some_id", [], ["test1", "test2"], ["test1", "test2"]),
            ("some_id", [{"some_id": []}], ["test1", "test2"], ["test1", "test2"]),
            ("some_id", [{"some_id": ["test0", "test3"]}], ["test1", "test2"], ["test0", "test1", "test2", "test3"]),
            ("some_id", [{"some_id": ["test1", "test2"]}], ["test1", "test2"], ["test1", "test2"]),
        ]
    )
    def test_merge_tags(self, entity_id, entity_tags, new_tags, expected_result_tags):
        result_tags = ClassWithTagsSupportMixin.merge_tags(entity_id, entity_tags, new_tags)
        assert result_tags == expected_result_tags

    @pytest.mark.parametrize(
        "entity_id, entity_tags, tags_to_remove, expected_result_tags",
        [
            ("some_id", [], ["test1", "test2"], []),
            ("some_id", [{"some_id": []}], ["test1", "test2"], []),
            ("some_id", [{"some_id": ["test1"]}], ["test1", "test2"], []),
            ("some_id", [{"some_id": ["test1", "test3"]}], ["test1", "test2"], ["test3"]),
        ]
    )
    def test_diff_tags(self, entity_id, entity_tags, tags_to_remove, expected_result_tags):
        result_tags = ClassWithTagsSupportMixin.diff_tags(entity_id, entity_tags, tags_to_remove)
        assert result_tags == expected_result_tags



from .clusters import ClustersClient
from .dataset_client import DatasetsClient
from .dataset_tag_client import DatasetTagsClient
from .dataset_version_client import DatasetVersionsClient
from .deployment_client import DeploymentsClient
from .experiment_client import ExperimentsClient
from .hyperparameter_client import HyperparameterJobsClient
from .job_client import JobsClient
from .machine_types_client import MachineTypesClient
from .machines_client import MachinesClient
from .model_client import ModelsClient
from .notebook_client import NotebooksClient
from .project_client import ProjectsClient
from .secret_client import SecretsClient
from .storage_provider_client import StorageProvidersClient
from .sdk_client import SdkClient
from .tensorboards_client import TensorboardClient
from .workflow_client import WorkflowsClient



import copy

from .. import logger as sdk_logger
from ..repositories.common import BaseRepository
from ..repositories.tags import ListTagRepository, UpdateTagRepository


class BaseClient(object):
    def __init__(
            self,
            api_key,
            ps_client_name=None,
            logger=sdk_logger.MuteLogger()
    ):
        """
        Base class. All client classes inherit from it.

        :param str api_key: your API key
        :param str ps_client_name:
        :param sdk_logger.Logger logger:
        """
        self.api_key = api_key
        self.ps_client_name = ps_client_name
        self.logger = logger

    def build_repository(self, repository_class, *args, **kwargs):
        """
        :param type[BaseRepository] repository_class:
        :rtype: BaseRepository
        """

        if self.ps_client_name is not None and kwargs.get("ps_client_name") is None:
            kwargs = copy.deepcopy(kwargs)
            kwargs["ps_client_name"] = self.ps_client_name

        repository = repository_class(*args, api_key=self.api_key, logger=self.logger, **kwargs)
        return repository


class TagsSupportMixin(object):
    entity = ""

    @staticmethod
    def merge_tags(entity_id, entity_tags, new_tags):
        result_tags = []
        if entity_tags:
            entity_tags = entity_tags[0].get(entity_id, [])

            result_tags = entity_tags + new_tags
        else:
            result_tags += new_tags
        return sorted(list(set(result_tags)))

    @staticmethod
    def diff_tags(entity_id, entity_tags, tags_to_remove):
        result_tags = []
        if entity_tags:
            entity_tags = entity_tags[0].get(entity_id, [])
            entity_tags = set(entity_tags) - set(tags_to_remove)
            result_tags = sorted(list(entity_tags))

        return result_tags

    def add_tags(self, entity_id, tags):
        """
        Add tags to entity.
        :param entity_id:
        :param entity:
        :param tags:
        :return:
        """
        list_tag_repository = self.build_repository(ListTagRepository)
        entity_tags = list_tag_repository.list(entity=self.entity, entity_ids=[entity_id])

        if entity_tags:
            tags = self.merge_tags(entity_id, entity_tags, tags)

        update_tag_repository = self.build_repository(UpdateTagRepository)
        update_tag_repository.update(entity=self.entity, entity_id=entity_id, tags=tags)

    def remove_tags(self, entity_id, tags):
        """
        Remove tags from entity.
        :param str entity_id:
        :param list[str] tags: list of tags to remove from entity
        :return:
        """
        list_tag_repository = self.build_repository(ListTagRepository)
        entity_tags = list_tag_repository.list(entity=self.entity, entity_ids=[entity_id])

        if entity_tags:
            entity_tags = self.diff_tags(entity_id, entity_tags, tags)

            update_tag_repository = self.build_repository(UpdateTagRepository)
            update_tag_repository.update(entity=self.entity, entity_id=entity_id, tags=entity_tags)

    def list_tags(self, entity_ids):
        """
        List tags for entity
        :param list[str] entity_ids:
        :return:
        """
        list_tag_repository = self.build_repository(ListTagRepository)
        entity_tags = list_tag_repository.list(entity=self.entity, entity_ids=entity_ids)
        return entity_tags



import attr


@attr.s
class VolumeOptions(object):
    kind = attr.ib(type=str, default=None)
    size = attr.ib(type=str, default=None)


@attr.s
class ExperimentDataset(object):
    id = attr.ib(type=str, default=None)
    uri = attr.ib(type=str, default=None)
    aws_access_key_id = attr.ib(type=str, default=None)
    aws_secret_access_key = attr.ib(type=str, default=None)
    aws_endpoint = attr.ib(type=str, default=None)
    etag = attr.ib(type=str, default=None)
    version_id = attr.ib(type=str, default=None)
    name = attr.ib(type=str, default=None)
    volume_options = attr.ib(type=VolumeOptions, factory=VolumeOptions)



from gradient import version_checker
from gradient.cli.cli import cli as _cli_entry_point


def main():
    version_checker.GradientVersionChecker.look_for_new_version_with_timeout()
    _cli_entry_point()



import click

from gradient.api_sdk import constants
from gradient.cli import common
from gradient.cli.cli import cli
from gradient.cli.cli_types import ChoiceType, json_string
from gradient.cli.common import validate_comma_split_option
from gradient.commands import models as models_commands


@cli.group("models", help="Manage models", cls=common.ClickGroup)
def models_group():
    pass


@models_group.group("tags", help="Manage model tags", cls=common.ClickGroup)
def model_tags():
    pass


@models_group.command("list", help="List models with optional filtering")
@click.option(
    "--experimentId",
    "experiment_id",
    help="Use to filter by experiment ID",
    cls=common.GradientOption,
)
@click.option(
    "--projectId",
    "project_id",
    help="Use to filter by project ID",
    cls=common.GradientOption,
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    cls=common.GradientOption,
    help="Filter by tags. Multiple use"
)
@common.api_key_option
@common.options_file
def list_models(api_key, experiment_id, project_id, tags, options_file):
    command = models_commands.ListModelsCommand(api_key=api_key)
    command.execute(experiment_id=experiment_id, project_id=project_id, tags=tags)


@models_group.command("delete", help="Delete model")
@click.option(
    "--id",
    "model_id",
    required=True,
    help="Model ID",
    cls=common.GradientOption,
)
@common.api_key_option
@common.options_file
def delete_model(api_key, model_id, options_file):
    command = models_commands.DeleteModelCommand(api_key=api_key)
    command.execute(model_id=model_id)


@models_group.command("create", help="Create a model from an url or dataset id")
@click.option(
    "--name",
    "name",
    required=True,
    help="Model name",
    cls=common.GradientOption,
)
@click.option(
    "--modelType",
    "model_type",
    required=True,
    type=ChoiceType(constants.MODEL_TYPES_MAP, case_sensitive=False),
    help="Model type",
    cls=common.GradientOption,
)
@click.option(
    "--datasetRef",
    "dataset_ref",
    required=True,
    help="Dataset ref to associate a model with",
    cls=common.GradientOption,
)
@click.option(
    "--projectId",
    "project_id",
    help="ID of a project",
    cls=common.GradientOption,
)
@click.option(
    "--modelSummary",
    "model_summary",
    type=json_string,
    help="Model summary",
    cls=common.GradientOption,
)
@click.option(
    "--notes",
    "notes",
    help="Additional notes",
    cls=common.GradientOption,
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to add to experiment",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want add to experiment",
    cls=common.GradientOption
)
@common.api_key_option
@common.options_file
def create_model(api_key, options_file, **model):
    model["tags"] = validate_comma_split_option(model.pop("tags_comma"), model.pop("tags"))
    command = models_commands.CreateModel(api_key=api_key)
    command.execute(**model)


@models_group.command("upload", help="Upload a model file or directory")
@click.argument(
    "PATH",
    required=True,
    type=click.Path(exists=True),
    cls=common.GradientArgument,
)
@click.option(
    "--name",
    "name",
    required=True,
    help="Model name",
    cls=common.GradientOption,
)
@click.option(
    "--modelType",
    "model_type",
    required=True,
    type=ChoiceType(constants.MODEL_TYPES_MAP, case_sensitive=False),
    help="Model type",
    cls=common.GradientOption,
)
@click.option(
    "--projectId",
    "project_id",
    help="ID of a project",
    cls=common.GradientOption,
)
@click.option(
    "--clusterId",
    "cluster_id",
    help="ID of a cluster",
    cls=common.GradientOption,
)
@click.option(
    "--modelSummary",
    "model_summary",
    type=json_string,
    help="Model summary",
    cls=common.GradientOption,
)
@click.option(
    "--notes",
    "notes",
    help="Additional notes",
    cls=common.GradientOption,
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to add to experiment",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want add to experiment",
    cls=common.GradientOption
)
@common.api_key_option
@common.options_file
def upload_model(api_key, options_file, **model):
    model["tags"] = validate_comma_split_option(model.pop("tags_comma"), model.pop("tags"))
    command = models_commands.UploadModel(api_key=api_key)
    command.execute(**model)


@models_group.command("details", help="Show model details")
@click.option(
    "--id",
    "model_id",
    required=True,
    help="Model ID",
    cls=common.GradientOption,
)
@common.api_key_option
@common.options_file
def model_details(model_id, api_key, options_file):
    command = models_commands.GetModelCommand(api_key=api_key)
    command.execute(model_id)


@models_group.command("download", help="Download model files")
@click.option(
    "--id",
    "model_id",
    required=True,
    help="Model ID",
    cls=common.GradientOption,
)
@click.option(
    "--destinationDir",
    "destination_directory",
    required=True,
    help="Destination directory",
    cls=common.GradientOption,
)
@common.api_key_option
@common.options_file
def download_model_files(model_id, destination_directory, api_key, options_file):
    command = models_commands.DownloadModelFiles(api_key=api_key)
    command.execute(model_id, destination_directory)


@model_tags.command("add", help="Add tags to ml model")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the model",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to add to ml model",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want add to ml model",
    cls=common.GradientOption
)
@common.api_key_option
@common.options_file
def ml_model_add_tag(id, options_file, api_key, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"), raise_if_no_values=True)

    command = models_commands.MLModelAddTagsCommand(api_key=api_key)
    command.execute(id, **kwargs)


@model_tags.command("remove", help="Remove tags from ml model")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the model",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to remove from ml model",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want to remove from ml model",
    cls=common.GradientOption
)
@common.api_key_option
@common.options_file
def ml_model_remove_tags(id, options_file, api_key, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"), raise_if_no_values=True)

    command = models_commands.MLModelRemoveTagsCommand(api_key=api_key)
    command.execute(id, **kwargs)



import os
import tempfile

from . import s3_uploader, archivers, utils, sdk_exceptions
from .logger import MuteLogger


class WorkspaceHandler(object):
    WORKSPACE_ARCHIVER_CLS = archivers.ZipArchiver

    def __init__(self, logger_=None, archiver_cls=None):
        """

        :param logger_: gradient.logger
        """
        self.logger = logger_ or MuteLogger()
        self.archiver_cls = archiver_cls or self.WORKSPACE_ARCHIVER_CLS

    def _zip_workspace(self, workspace_path, ignore_files):
        zip_file_name = 'workspace.zip'
        zip_file_path = os.path.join(tempfile.gettempdir(), zip_file_name)

        if ignore_files:
            ignore_files = ignore_files.split(",")
            ignore_files = [f.strip() for f in ignore_files]

        zip_archiver = self._get_workspace_archiver()
        zip_archiver.archive(workspace_path, zip_file_path, exclude=ignore_files)

        return zip_file_path

    def handle(self, input_data):
        workspace_path = input_data.get('workspace')
        if workspace_path is None:
            return None

        ignore_files = input_data.get('ignore_files')

        if utils.PathParser.is_remote_path(workspace_path):
            return workspace_path  # nothing to do

        if utils.PathParser.is_local_path(workspace_path):
            return self._handle_local_path(workspace_path, ignore_files)

        raise sdk_exceptions.WrongPathError("Invalid workspace path: {}".format(workspace_path))

    def _handle_local_path(self, workspace_path, ignore_files):
        workspace_path = os.path.abspath(workspace_path)
        if utils.PathParser.parse_path(workspace_path) == utils.PathParser.LOCAL_DIR:
            self.logger.log('Archiving your working directory for upload as your experiment workspace...'
                            '(See https://docs.paperspace.com/gradient/experiments/run-experiments for more '
                            'information.)')
            workspace_path = self._zip_workspace(workspace_path, ignore_files)

        return workspace_path

    def _get_workspace_archiver(self):
        workspace_archiver = self.archiver_cls(logger=self.logger)
        return workspace_archiver


class S3WorkspaceHandler(WorkspaceHandler):
    WORKSPACE_UPLOADER_CLS = s3_uploader.ExperimentFileUploader

    def __init__(self, api_key, client_name=None, uploader_cls=None, *args, **kwargs):
        """
        :param str api_key:
        :param str client_name:
        :param object uploader_cls:
        :param gradient.logger logger_:
        """
        super(S3WorkspaceHandler, self).__init__(*args, **kwargs)
        self.api_key = api_key
        self.client_name = client_name
        self.uploader_cls = uploader_cls or self.WORKSPACE_UPLOADER_CLS

    def handle(self, input_data):
        workspace = super(S3WorkspaceHandler, self).handle(input_data)
        if workspace is None:
            return None

        if utils.PathParser.is_remote_path(workspace):
            return workspace

        project_handle = input_data.get('projectHandle') or input_data.get("project_id")
        cluster_id = input_data.get('clusterId') or input_data.get("cluster_id")
        workspace = self._upload(workspace, project_handle, cluster_id=cluster_id)
        return workspace

    def _upload(self, archive_path, project_id, cluster_id=None):
        uploader = self._get_workspace_uploader(self.api_key)
        workspace = uploader.upload(archive_path, project_id, cluster_id=cluster_id)
        return workspace

    def _get_workspace_uploader(self, api_key):
        workspace_uploader = self.WORKSPACE_UPLOADER_CLS(api_key, logger=self.logger, ps_client_name=self.client_name)
        return workspace_uploader


class S3WorkspaceHandlerWithProgressbar(S3WorkspaceHandler):
    WORKSPACE_ARCHIVER_CLS = archivers.ZipArchiverWithProgressbar

    def _get_workspace_uploader(self, api_key):
        file_uploader = s3_uploader.S3FileUploader(
            s3_uploader.MultipartEncoderWithProgressbar,
            logger=self.logger,
            ps_client_name=self.client_name,
        )
        workspace_uploader = self.uploader_cls(
            api_key,
            uploader=file_uploader,
            logger=self.logger,
            ps_client_name=self.client_name
        )
        return workspace_uploader



import click

from gradient.cli import common
from gradient.cli.cli import cli
from gradient.cli.common import ClickGroup, api_key_option
from gradient.commands.workflows import ListWorkflowsCommand, ListWorkflowRunsCommand, GetWorkflowCommand, GetWorkflowRunCommand, CreateWorkflowCommand, CreateWorkflowRunCommand, WorkflowLogsCommand


@cli.group("workflows", help="Manage workflows", cls=ClickGroup)
def workflows():
    pass

@workflows.command("create", help="Create workflow")
@click.option(
    "--name",
    "name",
    required=True,
    help="Workflow name",
    cls=common.GradientOption,
)
@click.option(
    "--projectId",
    "project_id",
    required=True,
    help="Workflow name",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def create_workflow(api_key, name, project_id, options_file):
    command = CreateWorkflowCommand(api_key=api_key)
    command.execute(name=name, project_id=project_id)


@workflows.command("run", help="Run workflow spec")
@click.option(
    "--id",
    "workflow_id",
    required=True,
    help="Workflow ID",
    cls=common.GradientOption,
)
@click.option(
    "--clusterId",
    "cluster_id",
    required=False,
    help="Cluster ID",
    cls=common.GradientOption,
)
@click.option(
    "--path",
    "spec_path",
    required=True,
    help="Path to spec",
    cls=common.GradientOption,
)
@click.option(
    "--inputPath",
    "input_path",
    required=False,
    help="Path to inputs",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
@click.pass_context
def create_workflow_run(ctx, api_key, workflow_id, cluster_id, spec_path, input_path, options_file):
    command = CreateWorkflowRunCommand(api_key=api_key)
    workflow_run = command.execute(spec_path=spec_path, input_path=input_path, workflow_id=workflow_id, cluster_id=cluster_id)
    try:
        logId = workflow_run['status']['logId']
        # disable broken workflow-logs until we implement a functional job-specific or job-concatted log system
        # ctx.invoke(list_logs, workflow_log_id=logId, line=1, limit=100, follow=True, api_key=api_key)
    except KeyError:
        pass


@workflows.command("list", help="List workflows")
@click.option(
    "--projectId",
    "project_id",
    required=False,
    help="Project ID",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def get_workflows_list(api_key, project_id, options_file):
    command = ListWorkflowsCommand(api_key=api_key)
    command.execute(project_id=project_id)


@workflows.command("get", help="Get workflow info")
@click.option(
    "--id",
    "workflow_id",
    required=True,
    cls=common.GradientOption,
    help="Workflow Id",
)
@click.option(
    "--show-runs",
    "show_runs",
    prompt=False,
    required=False,
    is_flag=True,
    default=False,
    help="Fetch runs",
    cls=common.GradientOption,
)
@click.option(
    "--run",
    "run",
    prompt=False,
    required=False,
    help="Specify workload run",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def get_workflow(api_key, workflow_id, show_runs, run, options_file):
    if show_runs:
        if run is None:
            command = ListWorkflowRunsCommand(api_key=api_key)
            command.execute(workflow_id=workflow_id)
        else:
            command = GetWorkflowRunCommand(api_key=api_key)
            command.execute(workflow_id=workflow_id, run=run)
    else:
        command = GetWorkflowCommand(api_key=api_key)
        command.execute(workflow_id)



@workflows.command("logs", help="List logs for specific workflow")
@click.option(
    "--id",
    "workflow_id",
    help="Workflow ID",
    required=True,
    cls=common.GradientOption,
)
@click.option(
    "--run",
    "run",
    help="Specify workload run",
    required=True,
    cls=common.GradientOption,
)
@click.option(
    "--logId",
    "workflow_log_id",
    required=False,
    cls=common.GradientOption,
)
@click.option(
    "--line",
    "line",
    required=False,
    default=0,
    cls=common.GradientOption,
)
@click.option(
    "--limit",
    "limit",
    required=False,
    default=10000,
    cls=common.GradientOption,
)
@click.option(
    "--follow",
    "follow",
    required=False,
    default=False,
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
@click.pass_context
def list_logs(ctx, api_key, workflow_id, workflow_log_id, run, line, limit, follow, options_file):
    command = WorkflowLogsCommand(api_key=api_key)
    if workflow_log_id:
        command.execute(workflow_log_id, line, limit, follow)
    else:
        getRunCommand = GetWorkflowRunCommand(api_key=api_key)
        workflow_run = getRunCommand.get_instance(workflow_id, run)
        try:
            logId = workflow_run['status']['logId']
            command.execute(logId, line, limit, follow)
        except KeyError:
            pass




from .common import AlterResource, DeleteResource, GetResource, ListResources
from .datasets import DatasetMixin
from .. import serializers


class DatasetTagMixin(DatasetMixin):
    SERIALIZER_CLS = serializers.DatasetTagSchema

    @classmethod
    def get_request_url(cls, id='', name=None, **kwargs):
        dataset_id, _, ref = id.partition(':')
        if not name and ref:
            name = ref

        url = super(DatasetTagMixin, cls).get_request_url(id=dataset_id, **kwargs) + "/tags"
        if name:
            url += "/{}".format(name)
        return url


class ListDatasetTags(DatasetTagMixin, ListResources):
    def _get_request_params(self, kwargs):
        limit = kwargs.get("limit") or 20
        offset = kwargs.get("offset") or 0
        return {
            "filter[limit]": limit,
            "filter[skip]": offset,
            "filter[order][]": "name ASC",
        }


class SetDatasetTag(DatasetTagMixin, AlterResource):
    HANDLE_FIELD = "tag"

    def _get_request_json(self, instance):
        _, _, version = instance['id'].partition(':')
        instance = dict(instance)
        instance.pop('id', None)
        instance.pop('name', None)
        instance['version'] = version
        return instance

    def _send_request(self, client, url, json_data=None):
        response = client.put(url, json=json_data)
        return response


class GetDatasetTag(DatasetTagMixin, GetResource):
    pass


class DeleteDatasetTag(DatasetTagMixin, DeleteResource):
    pass



import re

import click

from gradient.api_sdk import constants

REQUIRED_PARAMS_PER_EXPERIMENT_TYPE = {
    constants.ExperimentType.GRPC_MULTI_NODE: ["parameter_server_container", "parameter_server_machine_type",
                                               "parameter_server_command", "parameter_server_count"],
    constants.ExperimentType.MPI_MULTI_NODE: ["master_container", "master_machine_type", "master_command",
                                              "master_count"],
}


def validate_mutually_exclusive(options_1, options_2, error_message):
    used_option_in_options_1 = any(option is not None for option in options_1)
    used_option_in_options_2 = any(option is not None for option in options_2)
    if used_option_in_options_1 and used_option_in_options_2:
        raise click.UsageError(error_message)


def validate_email(ctx, param, value):
    if value is not None \
            and not re.match(r"[^@]+@[^@]+\.[^@]+", value):
        raise click.BadParameter("Bad email address format")

    return value


def validate_multi_node(params):
    experiment_type = params.get('experiment_type_id')
    required_params = REQUIRED_PARAMS_PER_EXPERIMENT_TYPE.get(experiment_type)

    for param_name in required_params:
        if not params.get(param_name):
            raise click.UsageError("Param %s is required for this experiment type" % param_name)



from datetime import datetime

import attr


@attr.s
class Pagination(object):
    """
    Pagination class

    :param list[Any] data:
    :param str start_after:
    """
    data = attr.ib(type=list, default=None)
    start_after = attr.ib(type=str, default=None)


from .clusters import ListClusters
from .datasets import ListDatasets, CreateDataset, DeleteDataset, GetDataset, GetDatasetRef, UpdateDataset
from .dataset_tags import ListDatasetTags, GetDatasetTag, SetDatasetTag, DeleteDatasetTag
from .dataset_versions import ListDatasetVersions, CreateDatasetVersion, DeleteDatasetVersion, \
    GenerateDatasetVersionPreSignedS3Urls, GetDatasetVersion, UpdateDatasetVersion
from .deployments import ListDeployments, CreateDeployment, StartDeployment, StopDeployment, DeleteDeployment, \
    UpdateDeployment, GetDeployment, GetDeploymentMetrics, ListDeploymentMetrics, StreamDeploymentMetrics, ListDeploymentLogs
from .experiments import ListExperiments, GetExperiment, ListExperimentLogs, StartExperiment, StopExperiment, \
    CreateSingleNodeExperiment, CreateMultiNodeExperiment, RunSingleNodeExperiment, RunMultiNodeExperiment, \
    CreateMpiMultiNodeExperiment, RunMpiMultiNodeExperiment, DeleteExperiment, GetExperimentMetrics, ListExperimentMetrics, \
    StreamExperimentMetrics
from .hyperparameter import CreateHyperparameterJob, CreateAndStartHyperparameterJob, ListHyperparameterJobs, \
    GetHyperparameterTuningJob, StartHyperparameterTuningJob
from .jobs import ListJobs, ListResources, ListJobArtifacts, ListJobLogs, GetJob, GetJobMetrics, ListJobMetrics, StreamJobMetrics
from .machine_types import ListMachineTypes
from .machines import CheckMachineAvailability, CreateMachine, CreateResource, StartMachine, StopMachine, \
    RestartMachine, GetMachine, UpdateMachine, GetMachineUtilization
from .models import DeleteModel, ListModels, UploadModel, GetModel, ListModelFiles, CreateModel
from .notebooks import CreateNotebook, DeleteNotebook, GetNotebook, ListNotebooks, GetNotebookMetrics, ListNotebookMetrics, \
    StreamNotebookMetrics, StopNotebook, StartNotebook, ForkNotebook, ListNotebookArtifacts, ListNotebookLogs
from .projects import CreateProject, ListProjects, DeleteProject, GetProject
from .secrets import ListSecrets, SetSecret, DeleteSecret, EphemeralSecret
from .storage_providers import ListStorageProviders, CreateStorageProvider, DeleteStorageProvider, \
    GetStorageProvider, UpdateStorageProvider
from .tensorboards import CreateTensorboard, GetTensorboard, ListTensorboards, UpdateTensorboard, DeleteTensorboard
from .workflows import ListWorkflows, GetWorkflow, ListWorkflowRuns, GetWorkflowRun, CreateWorkflow, CreateWorkflowRun, ListWorkflowLogs



from datetime import datetime

import attr


@attr.s
class Project(object):
    """
    Project class

    :param str id:
    :param str name:
    :param str repository_name:
    :param str repository_url:
    :param datetime created:
    """
    id = attr.ib(type=str, default=None)
    name = attr.ib(type=str, default=None)
    repository_name = attr.ib(type=str, default=None)
    repository_url = attr.ib(type=str, default=None)
    created = attr.ib(type=datetime, default=None)
    tags = attr.ib(type=list, factory=list)



import abc
import os

import six

from gradient import api_sdk
from gradient.cli_constants import CLI_PS_CLIENT_NAME
from gradient.commands.common import BaseCommand, DetailsCommandMixin, ListCommandPagerMixin


def format_config(config, limit=None):
    if not config:
        return ''

    values = []

    for i, (name, value) in enumerate(sorted(config.items())):
        if not isinstance(value, six.string_types + (int, float, bool)):
            continue

        if limit is not None and i >= limit:
            values.append('...')
            break

        values.append("{}: {}".format(name, value))

    return os.linesep.join(values)


@six.add_metaclass(abc.ABCMeta)
class BaseStorageProvidersCommand(BaseCommand):
    def _get_client(self, api_key, logger):
        client = api_sdk.clients.StorageProvidersClient(
            api_key=api_key,
            logger=logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )
        return client


class ListStorageProvidersCommand(ListCommandPagerMixin, BaseStorageProvidersCommand):
    def _get_table_data(self, objects):
        data = [("Name", "ID", "Type", "Config")]
        for storage_provider in objects:
            data.append((
                storage_provider.name,
                storage_provider.id,
                storage_provider.type,
                format_config(storage_provider.config, limit=5)
            ))
        return data


class ShowStorageProviderDetailsCommand(DetailsCommandMixin, BaseStorageProvidersCommand):
    def _get_table_data(self, instance):
        data = (
            ("Name", instance.name),
            ("ID", instance.id),
            ("Type", instance.type),
            ("Config", format_config(instance.config))
        )
        return data


class CreateS3StorageProviderCommand(BaseStorageProvidersCommand):
    def execute(self, name, bucket, access_key, secret_access_key, endpoint=None, region=None, signature_version=None):
        storage_provider_id = self.client.create_s3(
            name=name,
            bucket=bucket,
            access_key=access_key,
            secret_access_key=secret_access_key,
            endpoint=endpoint,
            region=region,
            signature_version=signature_version,
        )
        self.logger.log("Created new storage provider with id: {}".format(storage_provider_id))


class UpdateS3StorageProviderCommand(BaseStorageProvidersCommand):
    def execute(self, storage_provider_id, name=None, bucket=None, access_key=None, secret_access_key=None,
                endpoint=None, region=None, signature_version=None):
        self.client.update_s3(
            storage_provider_id=storage_provider_id,
            name=name,
            bucket=bucket,
            access_key=access_key,
            secret_access_key=secret_access_key,
            endpoint=endpoint,
            region=region,
            signature_version=signature_version,
        )
        self.logger.log("Updated storage provider")


class DeleteStorageProviderCommand(BaseStorageProvidersCommand):
    def execute(self, storage_provider_id):
        self.client.delete(storage_provider_id)
        self.logger.log("Deleted storage provider: {}".format(storage_provider_id))



import copy

import marshmallow

from .experiment import BaseExperimentSchema
from .tag import TagSchema
from .. import models, utils


class HyperparameterSchema(BaseExperimentSchema):
    MODEL = models.Hyperparameter

    tuning_command = marshmallow.fields.Str(dump_to="tuningCommand", load_from="tuningCommand")
    worker_container = marshmallow.fields.Str(dump_to="workerContainer", load_from="workerContainer")
    worker_container_user = marshmallow.fields.Str(dump_to="workerContainerUser", load_from="workerContainerUser")
    worker_machine_type = marshmallow.fields.Str(dump_to="workerMachineType", load_from="workerMachineType")
    worker_command = marshmallow.fields.Str(dump_to="workerCommand", load_from="workerCommand")
    worker_count = marshmallow.fields.Int(dump_to="workerCount", load_from="workerCount")
    use_dockerfile = marshmallow.fields.Bool(dump_to="useDockerfile", load_from="useDockerfile")
    dockerfile_path = marshmallow.fields.Str(dump_to="dockerfilePath", load_from="dockerfilePath")
    worker_registry_username = marshmallow.fields.Str(dump_to="workerRegistryUsername")
    worker_registry_password = marshmallow.fields.Str(dump_to="workerRegistryPassword")

    hyperparameter_server_registry_username = marshmallow.fields.Str(dump_to="hyperparameterServerRegistryUsername",
                                                                     load_from="hyperparameterServerRegistryUsername")
    hyperparameter_server_registry_password = marshmallow.fields.Str(dump_to="hyperparameterServerRegistryPassword",
                                                                     load_from="hyperparameterServerRegistryPassword")
    hyperparameter_server_machine_type = marshmallow.fields.Str(dump_to="hyperparameterServerMachineType",
                                                                load_from="hyperparameterServerMachineType")
    hyperparameter_server_container = marshmallow.fields.Str(dump_to="hyperparameterServerContainer",
                                                             load_from="hyperparameterServerContainer")
    hyperparameter_server_container_user = marshmallow.fields.Str(dump_to="hyperparameterServerContainerUser",
                                                                  load_from="hyperparameterServerContainerUser")
    tags = marshmallow.fields.Nested(TagSchema, only="name", many=True, load_only=True)

    @marshmallow.pre_dump
    def preprocess(self, data, **kwargs):
        data = copy.copy(data)

        utils.base64_encode_attribute(data, "worker_command")
        utils.base64_encode_attribute(data, "tuning_command")
        return data



import io
import os
import re
import sys
from codecs import open

from setuptools import setup, find_packages
from setuptools.command.install import install

with io.open("gradient/version.py", "rt", encoding="utf8") as f:
    version = re.search(r"version = \"(.*?)\"", f.read()).group(1)

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
try:
    import pypandoc

    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError, OSError):
    with open('README.md', encoding='utf-8') as f:
        long_description = f.read()


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        if '0.0.0' == version:
            sys.exit("Version unspecified")


setup(
    name='gradient',
    version=version,
    description='Paperspace Python',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/paperspace/gradient-cli',
    author='Paperspace Co.',
    author_email='info@paperspace.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='paperspace api development library',
    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'old_tests']),
    install_requires=[
        'requests[security]',
        'six',
        'click==7.1.2',
        'terminaltables',
        'click-didyoumean',
        'click-help-colors',
        'click-completion',
        'colorama==0.4.3',
        'requests-toolbelt',
        'progressbar2',
        'halo',
        'marshmallow<3.0',
        'attrs<=19',
        'PyYAML==5.*',
        'python-dateutil==2.*',
        'websocket-client==0.57.*',
        'gradient-utils>=0.1.2',
    ],
    entry_points={'console_scripts': [
        'gradient = gradient:main.main',
    ]},
    extras_require={
        "dev": [
            'tox',
            'pytest',
            'mock',
            'twine',
            'sphinx',
            'pathlib2;python_version<"3.4"',
            'sphinx-click',
            'recommonmark'
        ],
    },
    cmdclass={
        'verify': VerifyVersionCommand,
    },
)



import abc
import pydoc

import halo
import six
import terminaltables

import gradient.cli_constants
from gradient import api_sdk
from gradient.cliutils import get_terminal_lines
from . import BaseCommand, common


@six.add_metaclass(abc.ABCMeta)
class GetTensorboardClientCommandMixin(BaseCommand):
    def _get_client(self, api_key, logger):
        client = api_sdk.clients.TensorboardClient(
            api_key=api_key,
            logger=logger,
            ps_client_name=gradient.cli_constants.CLI_PS_CLIENT_NAME,
        )
        return client

    def _log_object(self, instance):
        table_str = self._make_table(instance)
        if len(table_str.splitlines()) > get_terminal_lines():
            pydoc.pager(table_str)
        else:
            self.logger.log(table_str)

        # experiment table
        experiment_table_str = self._make_exp_table(instance)
        if len(experiment_table_str.splitlines()) > get_terminal_lines():
            pydoc.pager(experiment_table_str)
        else:
            self.logger.log(experiment_table_str)

    def _make_table(self, instance):
        """
        :param api_sdk.Tensorboard:
        """
        data = self._get_table_data(instance)
        ascii_table = terminaltables.AsciiTable(data)
        table_string = ascii_table.table
        return table_string

    def _make_exp_table(self, instance):
        """
        :param api_sdk.Tensorboard:
        """
        data = self._get_exp_table_data(instance)
        ascii_table = terminaltables.AsciiTable(data)
        table_string = ascii_table.table
        return table_string

    @staticmethod
    def _get_table_data(instance):
        """
        :param api_sdk.Tensorboard instance:
        """
        data = (
            ("ID", instance.id),
            ("Image", instance.image),
            ("URL", instance.url),
            ("State", instance.state),
            # ("Instance type", instance.instance.type),    TODO: for now
            # ("Instance size", instance.instance.size),
            # ("Instance count", instance.instance.count),
        )
        return data

    @staticmethod
    def _get_exp_table_data(instance):
        """
        :param api_sdk.Tensorboard instance:
        """
        data = [["Experiments ID", "State"]]

        for e in instance.experiments:
            data.append([e.get("id"), e.get('state')])
        return data


class CreateTensorboardCommand(GetTensorboardClientCommandMixin, common.BaseCommand):
    SPINNER_MESSAGE = "Creating new tensorboard"

    def execute(self, **kwargs):
        with halo.Halo(text=self.SPINNER_MESSAGE, spinner="dots"):
            notebook_id = self.client.create(**kwargs)

        self.logger.log("Created new tensorboard with id: {}".format(notebook_id))


class GetTensorboardCommand(GetTensorboardClientCommandMixin, common.BaseCommand):
    WAITING_FOR_RESPONSE_MESSAGE = "Waiting for data..."

    def execute(self, id_):
        with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
            instance = self._get_instance(id_)

        self._log_object(instance)

    def _get_instance(self, id_):
        """
        :rtype: api_sdk.Tensorboard
        """
        instance = self.client.get(id_)
        return instance


class ListTensorboardsCommand(GetTensorboardClientCommandMixin, common.ListCommandMixin, common.BaseCommand):
    WAITING_FOR_RESPONSE_MESSAGE = "Waiting for data..."

    def _get_instances(self, kwargs):
        instances = self.client.list()
        return instances

    def _get_table_data(self, objects):
        # TODO later we need to add information about state
        data = [["ID", "URL", "STATE"]]
        for obj in objects:
            data.append([obj.id, obj.url, obj.state])
        return data


class AddExperimentToTensorboard(GetTensorboardClientCommandMixin, common.BaseCommand):
    SPINNER_MESSAGE = "Adding experiments to tensorboard"

    def execute(self, id, experiments):
        """
        :param str id:
        :param list[str] experiments:
        """
        with halo.Halo(text=self.SPINNER_MESSAGE, spinner="dots"):
            tensorboard = self.client.add_experiments(id, added_experiments=list(experiments))

        self._log_object(tensorboard)


class RemoveExperimentToTensorboard(GetTensorboardClientCommandMixin, common.BaseCommand):
    SPINNER_MESSAGE = "Removing experiments from tensorboard"

    def execute(self, id, **kwargs):
        with halo.Halo(text=self.SPINNER_MESSAGE, spinner="dots"):
            tensorboard = self.client.remove_experiments(id, removed_experiments=list(kwargs.get('experiments')))

        self._log_object(tensorboard)


class DeleteTensorboard(GetTensorboardClientCommandMixin, common.BaseCommand):
    SPINNER_MESSAGE = "Deleting tensorboard"

    def execute(self, *args, **kwargs):
        with halo.Halo(text=self.SPINNER_MESSAGE, spinner="dots"):
            self.client.delete(**kwargs)

        self.logger.log("Delete tensorboard ({}) ended with success".format(kwargs.get("id")))



import fnmatch
import os
import zipfile

import progressbar

from .logger import MuteLogger


class ZipArchiver(object):
    DEFAULT_EXCLUDED_PATHS = [
        os.path.join(".git", "*"),
        os.path.join(".idea", "*"),
        os.path.join(".pytest_cache", "*"),
    ]

    def __init__(self, logger=None):
        self.logger = logger or MuteLogger()
        self.default_excluded_paths = self.DEFAULT_EXCLUDED_PATHS[:]

    def archive(self, input_dir_path, output_file_path, overwrite_existing_archive=True, exclude=None):
        """

        :param str input_dir_path:
        :param str output_file_path:
        :param bool overwrite_existing_archive:
        :param list|tuple|None exclude:
        """
        excluded_paths = self.get_excluded_paths(exclude)

        file_paths = self.get_file_paths(input_dir_path, excluded_paths)

        if os.path.exists(output_file_path):
            if not overwrite_existing_archive:
                raise IOError("File already exists")

            self.logger.log('Removing existing archive')
            os.remove(output_file_path)

        self.logger.log('Creating zip archive: %s' % output_file_path)
        self._archive(file_paths, output_file_path)
        self.logger.log('Finished creating archive: %s' % output_file_path)

    def get_excluded_paths(self, exclude=None):
        """
        :param list|tuple|None exclude:
        :rtype: set
        """
        if exclude is None:
            exclude = []

        excluded_paths = set(self.default_excluded_paths)
        excluded_paths.update(exclude)
        return excluded_paths

    @staticmethod
    def get_file_paths(input_path, excluded_paths=None):
        """Get a dictionary of all files in input_dir excluding specified in excluded_paths

        :param str input_path:
        :param list|tuple|set|None excluded_paths:
        :return: dictionary with full paths as values as keys and relative paths
        :rtype: dict[str,str]
        """
        if excluded_paths is None:
            excluded_paths = []

        file_paths = {}

        # Read all directory, subdirectories and file lists
        for root, dirs, files in os.walk(input_path):
            relative_path = os.path.relpath(root, input_path)

            for filename in files:
                # Create the full filepath by using os module.
                if relative_path == '.':
                    file_path = filename
                else:
                    file_path = os.path.join(os.path.relpath(root, input_path), filename)

                if any(fnmatch.fnmatch(file_path, pattern) for pattern in excluded_paths):
                    continue

                if file_path not in excluded_paths:
                    file_paths[file_path] = os.path.join(root, filename)

        return file_paths

    def _archive(self, file_paths, output_file_path):
        """Create ZIP archive and add files to it

        :param dict[str,str] file_paths:
        :param str output_file_path:
        """
        zip_file = zipfile.ZipFile(output_file_path, 'w')
        with zip_file:
            i = 0
            for relative_path, abspath in file_paths.items():
                i += 1
                self.logger.debug('Adding %s to archive' % relative_path)
                zip_file.write(abspath, arcname=relative_path)
                self._archive_iterate_callback(i)

    def _archive_iterate_callback(self, i):
        pass


class ZipArchiverWithProgressbar(ZipArchiver):
    def _archive(self, file_paths, output_file_path):
        """Create ZIP archive and add files to it and show progress bar in terminal

        :param dict[str,str] file_paths:
        :param str output_file_path:
        """
        self.bar = progressbar.ProgressBar(max_value=len(file_paths))
        super(ZipArchiverWithProgressbar, self)._archive(file_paths, output_file_path)
        self.bar.finish()

    def _archive_iterate_callback(self, i):
        self.bar.update(i)



from .common import ListResources, CreateResource, StartResource, StopResource, DeleteResource, AlterResource, \
    GetResource, GetMetrics, ListMetrics, StreamMetrics, ListLogs
from .. import serializers, config, sdk_exceptions
from ..sdk_exceptions import ResourceFetchingError, MalformedResponseError


class GetBaseDeploymentApiUrlMixin(object):
    def _get_api_url(self, **_):
        return config.config.CONFIG_HOST


class ListDeployments(GetBaseDeploymentApiUrlMixin, ListResources):
    def get_request_url(self, **kwargs):
        return "/deployments/getDeploymentList/"

    def _parse_objects(self, data, **kwargs):
        deployment_dicts = self._get_deployments_dicts_from_json_data(data, kwargs)
        deployments = []

        for deployment_dict in deployment_dicts:
            deployment = serializers.DeploymentSchema().get_instance(deployment_dict)
            deployments.append(deployment)

        return deployments

    @staticmethod
    def _get_deployments_dicts_from_json_data(data, kwargs):
        return data["deploymentList"]

    def _get_request_json(self, kwargs):
        filters = {}
        if kwargs["model_id"]:
            filters["modelId"] = kwargs["model_id"]

        if kwargs["state"]:
            filters["state"] = kwargs["state"]

        if kwargs["project_id"]:
            filters["projectId"] = kwargs["project_id"]

        if filters:
            json_ = {"filter": {"where": {"and": [filters]}}}
        else:
            json_ = {}

        tags = kwargs.get("tags")
        if tags:
            json_["tagFilter"] = tags

        return json_ or None


class CreateDeployment(GetBaseDeploymentApiUrlMixin, CreateResource):
    SERIALIZER_CLS = serializers.DeploymentCreateSchema

    def get_request_url(self, **kwargs):
        if kwargs.get("cluster"):
            return "/deployments/v2/createDeployment/"

        return "/deployments/createDeployment/"

    def _get_id_from_response(self, response):
        handle = response.data["deployment"]["id"]
        return handle


class StartDeployment(GetBaseDeploymentApiUrlMixin, StartResource):
    def get_request_url(self, **kwargs):
        return "/deployments/v2/updateDeployment/"

    def _get_request_json(self, kwargs):
        data = {
            "id": kwargs["id"],
            "isRunning": True,
        }
        return data

    def _send_request(self, client, url, json_data=None):
        response = client.post(url, json=json_data)
        return response


class StopDeployment(GetBaseDeploymentApiUrlMixin, StopResource):
    def get_request_url(self, **kwargs):
        return "/deployments/v2/updateDeployment/"

    def _get_request_json(self, kwargs):
        data = {
            "id": kwargs["id"],
            "isRunning": False,
        }
        return data

    def _send_request(self, client, url, json_data=None):
        response = client.post(url, json=json_data)
        return response


class DeleteDeployment(GetBaseDeploymentApiUrlMixin, DeleteResource):
    def get_request_url(self, **kwargs):
        return "/deployments/v2/deleteDeployment"

    def _get_request_json(self, kwargs):
        data = {
            "id": kwargs["id"],
            "isRunning": False,
        }
        return data

    def _send_request(self, client, url, json_data=None):
        response = client.post(url, json=json_data)
        return response


class UpdateDeployment(GetBaseDeploymentApiUrlMixin, AlterResource):
    SERIALIZER_CLS = serializers.DeploymentSchema
    VALIDATION_ERROR_MESSAGE = "Failed to update resource"

    def update(self, id, instance):
        instance_dict = self._get_instance_dict(instance)
        self._run(id=id, **instance_dict)

    def get_request_url(self, **kwargs):
        return "/deployments/v2/updateDeployment"

    def _get_request_json(self, kwargs):
        # this temporary workaround is here because create and update
        # endpoints have different names for docker args field
        args = kwargs.pop("dockerArgs", None)
        if args:
            kwargs["args"] = args

        j = {
            "id": kwargs.pop("id"),
            "upd": kwargs,
        }
        return j


class GetDeployment(GetBaseDeploymentApiUrlMixin, GetResource):
    SERIALIZER_CLS = serializers.DeploymentSchema

    def get_request_url(self, **kwargs):
        return "/deployments/getDeploymentList/"

    def _get_request_json(self, kwargs):
        deployment_id = kwargs["deployment_id"]
        filter_ = {"where": {"and": [{"id": deployment_id}]}}
        json_ = {"filter": filter_}
        return json_

    def _parse_object(self, instance_dict, **kwargs):
        try:
            instance_dict = instance_dict["deploymentList"][0]
        except KeyError:
            raise MalformedResponseError("Malformed response from API")
        except IndexError:
            raise ResourceFetchingError("Deployment not found")

        return super(GetDeployment, self)._parse_object(instance_dict, **kwargs)


class GetDeploymentMetrics(GetMetrics):
    OBJECT_TYPE = "modelDeployment"

    def _get_instance_by_id(self, instance_id, **kwargs):
        repository = GetDeployment(self.api_key, logger=self.logger, ps_client_name=self.ps_client_name)
        instance = repository.get(deployment_id=instance_id)
        return instance

    def _get_start_date(self, instance, kwargs):
        rv = super(GetDeploymentMetrics, self)._get_start_date(instance, kwargs)
        if rv is None:
            raise sdk_exceptions.GradientSdkError("Deployment job has not started yet")

        return rv

class ListDeploymentMetrics(ListMetrics):
    OBJECT_TYPE = "modelDeployment"

    def _get_instance_by_id(self, instance_id, **kwargs):
        repository = GetDeployment(self.api_key, logger=self.logger, ps_client_name=self.ps_client_name)
        instance = repository.get(deployment_id=instance_id)
        return instance

    def _get_start_date(self, instance, kwargs):
        rv = super(ListDeploymentMetrics, self)._get_start_date(instance, kwargs)
        if rv is None:
            raise sdk_exceptions.GradientSdkError("Deployment job has not started yet")

        return rv

class StreamDeploymentMetrics(StreamMetrics):
    OBJECT_TYPE = "modelDeployment"

    def _get_metrics_api_url(self, instance_id, protocol="https"):
        repository = GetDeployment(api_key=self.api_key, logger=self.logger, ps_client_name=self.ps_client_name)
        deployment = repository.get(deployment_id=instance_id)

        metrics_api_url = super(StreamDeploymentMetrics, self)._get_metrics_api_url(deployment, protocol="wss")
        return metrics_api_url


class ListDeploymentLogs(ListLogs):
    def _get_request_params(self, kwargs):
        params = {
            "deploymentId": kwargs["id"],
            "line": kwargs["line"],
            "limit": kwargs["limit"]
        }
        return params



from .base_client import BaseClient, TagsSupportMixin
from .. import repositories, models


class NotebooksClient(TagsSupportMixin, BaseClient):
    entity = "notebook"

    def create(
            self,
            machine_type,
            project_id=None,
            cluster_id=None,
            container=None,
            container_id=None,
            name=None,
            registry_username=None,
            registry_password=None,
            command=None,
            container_user=None,
            shutdown_timeout=None,
            is_preemptible=None,
            is_public=None,
            workspace=None,
            workspace_ref=None,
            workspace_username=None,
            workspace_password=None,
            tags=None,
            environment=None,
    ):
        """Create new notebook

        :param str machine_type:
        :param int container_id:
        :param str|int project_id:
        :param str cluster_id:
        :param str container:
        :param str name:
        :param str registry_username:
        :param str registry_password:
        :param str command:
        :param str container_user:
        :param int shutdown_timeout:
        :param bool is_preemptible:
        :param bool is_public:
        :param list[str] tags: List of tags
        :param str workspace: Project git repository url
        :param str workspace_ref: Git commit hash, branch name or tag
        :param str workspace_username: Project git repository username
        :param str workspace_password: Project git repository password
        :param dict environment: key value collection of envs that are used in notebook

        :return: Notebook ID
        :rtype str:
        """

        notebook = models.Notebook(
            container_id=container_id,
            project_id=project_id,
            cluster_id=cluster_id,
            container=container,
            name=name,
            registry_username=registry_username,
            registry_password=registry_password,
            command=command,
            container_user=container_user,
            shutdown_timeout=shutdown_timeout,
            is_preemptible=is_preemptible,
            machine_type=machine_type,
            is_public=is_public,
            environment=environment,
            workspace=workspace,
            workspace_username=workspace_username,
            workspace_password=workspace_password,
            workspace_ref=workspace_ref,
        )

        repository = self.build_repository(repositories.CreateNotebook)
        handle = repository.create(notebook)

        if tags:
            self.add_tags(entity_id=handle, tags=tags)

        return handle

    def start(
            self,
            id,
            machine_type,
            cluster_id=None,
            name=None,
            shutdown_timeout=None,
            is_preemptible=None,
            tags=None,
    ):
        """Start existing notebook
        :param str|int id:
        :param str machine_type:
        :param str cluster_id:
        :param str name:
        :param int shutdown_timeout:
        :param bool is_preemptible:
        :param list[str] tags: List of tags

        :return: Notebook ID
        :rtype str:
        """
        notebook = models.NotebookStart(
            notebook_id=id,
            machine_type=machine_type,
            cluster_id=cluster_id,
            notebook_name=name,
            shutdown_timeout=shutdown_timeout,
            is_preemptible=is_preemptible,
        )

        repository = self.build_repository(repositories.StartNotebook)

        handle = repository.start(notebook)

        if tags:
            self.add_tags(entity_id=handle, tags=tags)

        return handle

    def fork(self, id, project_id, tags=None):
        """Fork an existing notebook
        :param str|int id:
        :param str project_id:
        :param list[str] tags: List of tags

        :return: Notebook ID
        :rtype str:
        """
        repository = self.build_repository(repositories.ForkNotebook)
        handle = repository.fork(id, project_id)

        if tags:
            self.add_tags(entity_id=handle, tags=tags)

        return handle

    def get(self, id):
        """Get Notebook

        :param str id: Notebook ID
        :rtype: models.Notebook
        """
        repository = self.build_repository(repositories.GetNotebook)
        notebook = repository.get(id=id)
        return notebook

    def delete(self, id):
        """Delete existing notebook

        :param str id: Notebook ID
        """
        repository = self.build_repository(repositories.DeleteNotebook)
        repository.delete(id)

    def list(self, tags=None, limit=None, offset=None, get_meta=False):
        """Get list of Notebooks

        :rtype: list[models.Notebook]
        """
        repository = self.build_repository(repositories.ListNotebooks)
        notebooks = repository.list(tags=tags, limit=limit, offset=offset, get_meta=get_meta)
        return notebooks

    def get_metrics(self, notebook_id, start=None, end=None, interval="30s", built_in_metrics=None):
        """Get notebook metrics

        :param str notebook_id: notebook ID
        :param datetime.datetime|str start:
        :param datetime.datetime|str end:
        :param str interval:
        :param list[str] built_in_metrics: List of metrics to get if different than default
                    Available builtin metrics: cpuPercentage, memoryUsage, gpuMemoryFree, gpuMemoryUsed, gpuPowerDraw,
                                            gpuTemp, gpuUtilization, gpuMemoryUtilization

        :returns: Metrics of a notebook
        :rtype: dict[str,dict[str,list[dict]]]
        """

        repository = self.build_repository(repositories.GetNotebookMetrics)
        metrics = repository.get(
            id=notebook_id,
            start=start,
            end=end,
            interval=interval,
            built_in_metrics=built_in_metrics,
        )
        return metrics

    def list_metrics(self, notebook_id, start=None, end=None, interval="30s"):
        """List notebook metrics

        :param str notebook_id: notebook ID
        :param datetime.datetime|str start:
        :param datetime.datetime|str end:
        :param str interval:
        :returns: Metrics of a notebook
        :rtype: dict[str,dict[str,list[dict]]]
        """

        repository = self.build_repository(repositories.ListNotebookMetrics)
        metrics = repository.get(
            id=notebook_id,
            start=start,
            end=end,
            interval=interval,
        )
        return metrics

    def stream_metrics(self, notebook_id, interval="30s", built_in_metrics=None):
        """Stream live notebook metrics

        :param str notebook_id: notebook ID
        :param str interval:
        :param list[str] built_in_metrics: List of metrics to get if different than default
                    Available builtin metrics: cpuPercentage, memoryUsage, gpuMemoryFree, gpuMemoryUsed, gpuPowerDraw,
                                            gpuTemp, gpuUtilization, gpuMemoryUtilization

        :returns: Generator object yielding live notebook metrics
        :rtype: Iterable[dict]
        """

        repository = self.build_repository(repositories.StreamNotebookMetrics)
        metrics = repository.stream(
            id=notebook_id,
            interval=interval,
            built_in_metrics=built_in_metrics,
        )
        return metrics

    def stop(self, id):
        """Stop existing notebook

        :param str|int id: Notebook ID
        """
        repository = self.build_repository(repositories.StopNotebook)
        repository.stop(id)

    def artifacts_list(self, notebook_id, files=None, size=False, links=True):
        """
        Method to retrieve all artifacts files.

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            artifacts = notebook_client.artifacts_list(
                notebook_id='your_notebook_id_here',
                files='your_files,here',
                size=False,
                links=True
            )

        :param str notebook_id: to limit artifact from this notebook.
        :param str files: to limit result only to file names provided. You can use wildcard option ``*``.
        :param bool size: flag to show file size. Default value is set to False.
        :param bool links: flag to show file url. Default value is set to True.

        :returns: list of files with description if specified from notebook artifacts.
        :rtype: list[Artifact]
        """
        repository = self.build_repository(repositories.ListNotebookArtifacts)
        artifacts = repository.list(notebook_id=notebook_id, files=files, links=links, size=size)
        return artifacts

    def logs(self, notebook_id, line=1, limit=10000):
        """
        Method to retrieve notebook logs.

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            notebook_logs = notebook_client.logs(
                notebook_id='Your_job_id_here',
                line=100,
                limit=100
            )

        :param str notebook_id: id of notebook that we want to retrieve logs
        :param int line: from what line you want to retrieve logs. Default 0
        :param int limit: how much lines you want to retrieve logs. Default 10000

        :returns: list of formatted logs lines
        :rtype: list
        """
        notebook = self.get(notebook_id)
        repository = self.build_repository(repositories.ListNotebookLogs)
        logs = repository.list(job_id=notebook.job_handle, notebook_id=notebook_id, line=line, limit=limit)
        return logs

    def yield_logs(self, notebook_id, line=1, limit=10000):
        """Get log generator. Polls the API for new logs

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            notebook_logs_generator = notebook_client.yield_logs(
                notebook_id='Your_job_id_here',
                line=100,
                limit=100
            )

        :param str notebook_id:
        :param int line: line number at which logs starts to display on screen
        :param int limit: maximum lines displayed on screen, default set to 10 000

        :returns: generator yielding LogRow instances
        :rtype: Iterator[models.LogRow]
        """

        notebook = self.get(notebook_id)
        repository = self.build_repository(repositories.ListNotebookLogs)
        logs = repository.yield_logs(job_id=notebook.job_handle, notebook_id=notebook_id, line=line, limit=limit)
        return logs



import abc

import six

from gradient import api_sdk
from gradient.cli_constants import CLI_PS_CLIENT_NAME
from gradient.commands.common import BaseCommand, ListCommandMixin


@six.add_metaclass(abc.ABCMeta)
class BaseMachineTypesCommand(BaseCommand):
    def _get_client(self, api_key, logger):
        client = api_sdk.clients.MachineTypesClient(
            api_key=api_key,
            logger=logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )
        return client


class ListMachineTypesCommand(ListCommandMixin, BaseMachineTypesCommand):
    def _get_instances(self, kwargs):
        instances = self.client.list(**kwargs)
        return instances

    def _get_table_data(self, objects):
        """
        :param list[VmTypeSchema] objects: object
        """
        data = [("Name", "Kind", "CPU Count", "RAM [Bytes]", "GPU Count", "GPU Model", "Clusters")]
        for obj in objects:
            name = obj.label
            kind = obj.kind
            cpu_count = obj.cpu_count
            ram_in_bytes = obj.ram_in_bytes
            gpu_count = obj.gpu_count
            gpu_model = obj.gpu_model.label if obj.gpu_model else "N/A"
            clusters = ", ".join(obj.clusters) if obj.clusters else ""

            data.append((name, kind, cpu_count, ram_in_bytes, gpu_count, gpu_model, clusters))

        return data



import datetime

import terminaltables

from gradient import api_sdk
from gradient.api_sdk.config import config
from gradient.api_sdk.utils import concatenate_urls
from gradient.cli_constants import CLI_PS_CLIENT_NAME
from gradient.commands import common, BaseCommand
from gradient.commands.common import DetailsCommandMixin


class GetMachinesClientMixin(object):
    def _get_client(self, api_key, logger):
        client = api_sdk.MachinesClient(
            api_key=api_key,
            logger=logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )
        return client


class CheckAvailabilityCommand(GetMachinesClientMixin, BaseCommand):
    def execute(self, machine_type, region):
        is_available = self.client.is_available(machine_type, region)
        self.logger.log("Machine available: {}".format(is_available))


class CreateMachineCommand(GetMachinesClientMixin, BaseCommand):
    def execute(self, kwargs):
        handle = self.client.create(**kwargs)
        self.logger.log("New machine created with id: {}".format(handle))
        self.logger.log(self.get_instance_url(handle))

    def get_instance_url(self, machine_id):
        url = concatenate_urls(config.WEB_URL, "{}/machines/{}".format(self.get_namespace(), machine_id))
        return url


class UpdateMachineCommand(GetMachinesClientMixin, BaseCommand):
    def execute(self, machine_id, kwargs):
        self.client.update(machine_id, **kwargs)
        self.logger.log("Machine updated")


class StartMachineCommand(GetMachinesClientMixin, BaseCommand):
    def execute(self, id):
        self.client.start(id)
        self.logger.log("Machine started")


class StopMachineCommand(GetMachinesClientMixin, BaseCommand):
    def execute(self, id):
        self.client.stop(id)
        self.logger.log("Machine stopped")


class RestartMachineCommand(GetMachinesClientMixin, BaseCommand):
    def execute(self, id):
        self.client.restart(id)
        self.logger.log("Machine restarted")


class ShowMachineCommand(DetailsCommandMixin, GetMachinesClientMixin, common.BaseCommand):
    def _get_table_data(self, machine):
        """
        :param api_sdk.Machine machine:
        """
        try:
            if machine.events:
                sorted_events = sorted(machine.events,
                                       key=lambda x: datetime.datetime.strptime(x.created, "%Y-%m-%dT%H:%M:%S.%fZ"))
                last_event = sorted_events[-1]
                last_event_string = "name:     {}\nstate:    {}\ncreated:  {}" \
                    .format(last_event.name, last_event.state, last_event.created)
            else:
                last_event_string = ""

        except (KeyError, IndexError):
            last_event_string = None

        data = (
            ("ID", machine.id),
            ("Name", machine.name),
            ("OS", machine.os),
            ("RAM", machine.ram),
            ("CPU", machine.cpus),
            ("GPU", machine.gpu),
            ("Storage Total", machine.storage_total),
            ("Storage Used", machine.storage_used),
            ("Usage Rate", machine.usage_rate),
            ("Shutdown Timeout In Hours", machine.shutdown_timeout_in_hours),
            ("Shutdown Timeout Forces", machine.shutdown_timeout_forces),
            ("Perform Auto Snapshot", machine.perform_auto_snapshot),
            ("Auto snapshot frequency", machine.auto_snapshot_frequency),
            ("Auto Snapshot Save Count", machine.auto_snapshot_save_count),
            ("Agent Type", machine.agent_type),
            ("Created", machine.created_timestamp),
            ("State", machine.state),
            ("Updates Pending", machine.updates_pending),
            ("Network ID", machine.network_id),
            ("Private IP Address", machine.private_ip_address),
            ("Public IP Address", machine.public_ip_address),
            ("Region", machine.region),
            ("Script ID", machine.script_id),
            ("Last Run", machine.last_run_timestamp),
            ("Dynamic Public IP", machine.dynamic_public_ip),
            ("Last event", last_event_string),
        )
        return data


class ListMachinesCommand(GetMachinesClientMixin, common.ListCommandMixin, BaseCommand):
    def _get_instances(self, kwargs):
        instances = self.client.list(**kwargs)
        return instances

    def _get_table_data(self, machines):
        """
        :param list[api_sdk.Machine] machines:
        :return:
        """
        data = [("ID", "Name", "OS", "CPU", "GPU", "RAM", "State", "Region")]
        for machine in machines:
            id_ = machine.id
            name = machine.name
            os_ = machine.os
            cpus = machine.cpus
            gpu = machine.gpu
            ram = machine.ram
            state = machine.state
            region = machine.region

            data.append((id_, name, os_, cpus, gpu, ram, state, region))

        return data

    def _get_request_json(self, kwargs):
        json_ = {"params": kwargs} if kwargs else None
        return json_


class DestroyMachineCommand(GetMachinesClientMixin, BaseCommand):
    def execute(self, machine_type, region):
        self.client.delete(machine_type, region)
        self.logger.log("Machine successfully destroyed")


class ShowMachineUtilizationCommand(GetMachinesClientMixin, BaseCommand):
    def execute(self, machine_id, billing_month):
        usage = self.client.get_utilization(machine_id, billing_month)
        table = self.make_details_table(usage)
        self.logger.log(table)

    @staticmethod
    def make_details_table(usage):
        data = (
            ("ID", usage.machine_id),
            ("Machine Seconds used", usage.machine_seconds_used),
            ("Machine Hourly rate", usage.machine_hourly_rate),
            ("Storage Seconds Used", usage.storage_seconds_used),
            ("Storage Monthly Rate", usage.storage_monthly_rate),
        )
        ascii_table = terminaltables.AsciiTable(data)
        table_string = ascii_table.table
        return table_string


class WaitForMachineStateCommand(GetMachinesClientMixin, BaseCommand):
    def execute(self, machine_id, state, interval=5):
        self.client.wait_for_state(machine_id, state, interval)
        self.logger.log("Machine state: {}".format(state))


class MachineAddTagsCommand(GetMachinesClientMixin, BaseCommand):
    def execute(self, machine_id, *args, **kwargs):
        self.client.add_tags(machine_id, **kwargs)
        self.logger.log("Tags added to machine")


class MachineRemoveTagsCommand(GetMachinesClientMixin, BaseCommand):
    def execute(self, machine_id, *args, **kwargs):
        self.client.remove_tags(machine_id, **kwargs)
        self.logger.log("Tags removed from machine")



import json

import gradient.api_sdk.config
from .common import ListResources, CreateResource, GetResource, DeleteResource, StopResource, GetMetrics, ListMetrics, StreamMetrics, \
    ListLogs
from .. import serializers, sdk_exceptions
from ..clients import http_client
from ..serializers import JobSchema


class GetBaseJobApiUrlMixin(object):
    def _get_api_url(self, **_):
        return gradient.api_sdk.config.config.CONFIG_HOST


class ListJobs(GetBaseJobApiUrlMixin, ListResources):

    def get_request_url(self, **kwargs):
        return "/jobs/getJobList/"

    def _parse_objects(self, data, **kwargs):
        jobs = []

        for job_dict in data["jobList"]:
            job = self._parse_object(job_dict)
            jobs.append(job)

        return jobs

    def _parse_object(self, job_dict):
        job = serializers.JobSchema().get_instance(job_dict)
        return job

    def _get_request_params(self, kwargs):
        filters = {"filter": {"where": {}}}
        if kwargs.get("project_id"):
            filters["filter"]["where"]["projectId"] = kwargs.get("project_id")

        if kwargs.get("project"):
            filters["filter"]["where"]["project"] = kwargs.get("project")

        if kwargs.get("experiment_id"):
            filters["filter"]["where"]["experimentId"] = kwargs.get("experiment_id")

        params = {}
        filter_string = json.dumps(filters)
        params["filter"] = filter_string

        tags = kwargs.get("tags")
        if tags:
            params["modelName"] = "team"  # TODO: filtering by tags won't work without this. Remove this when fixed.
            for i, tag in enumerate(tags):
                key = "tagFilter[{}]".format(i)
                params[key] = tag

        return params or None


class ListJobLogs(ListLogs):
    def _get_request_params(self, kwargs):
        params = {
            "jobId": kwargs["id"],
            "line": kwargs["line"],
            "limit": kwargs["limit"]
        }
        return params


class CreateJob(GetBaseJobApiUrlMixin, CreateResource):
    SERIALIZER_CLS = JobSchema
    HANDLE_FIELD = "id"

    def get_request_url(self, **kwargs):
        return "/jobs/createJob/"

    def _get_id_from_response(self, response):
        handle = response.data[self.HANDLE_FIELD]
        return handle

    def _get_request_json(self, instance_dict):
        return

    def _get_request_params(self, instance_dict):
        if instance_dict.get('datasets'):
            instance_dict['datasets'] = json.dumps(instance_dict['datasets'])

        return instance_dict


class DeleteJob(GetBaseJobApiUrlMixin, DeleteResource):

    def get_request_url(self, **kwargs):
        return "/jobs/{}/destroy".format(kwargs.get("id"))

    def _send_request(self, client, url, json_data=None):
        response = client.post(url, json=json_data)
        return response


class StopJob(GetBaseJobApiUrlMixin, StopResource):

    def get_request_url(self, **kwargs):
        return "/jobs/{}/stop".format(kwargs.get('id'))

    def _send_request(self, client, url, json_data=None):
        response = client.post(url, json=json_data)
        return response


class GetJob(GetBaseJobApiUrlMixin, GetResource):
    def get_request_url(self, **kwargs):
        return "/jobs/getPublicJob"

    def _get_request_json(self, kwargs):
        json_ = {
            "jobId": kwargs["job_id"]
        }
        return json_

    def _send_request(self, client, url, json=None, params=None):
        response = client.post(url, json=json, params=params)
        return response

    def _parse_object(self, instance_dict, **kwargs):
        instance_dict = instance_dict["job"]
        job = serializers.JobSchema().get_instance(instance_dict)
        return job


class ListJobArtifacts(GetBaseJobApiUrlMixin, ListResources):
    def _parse_objects(self, data, **kwargs):
        serializer = serializers.utils.paginate_schema(serializers.ArtifactSchema)
        return serializer.get_instance(data)

    def get_request_url(self, **kwargs):
        return "/jobs/artifactsListV2"

    def _get_request_params(self, kwargs):
        params = {
            "jobId": kwargs.get("jobId"),
        }

        if kwargs.get("files"):
            params["files"] = kwargs.get("files")

        if kwargs.get("size"):
            params["size"] = kwargs.get("size")

        if kwargs.get("links"):
            params["links"] = kwargs.get("links")

        if kwargs.get("start_after"):
            params["startAfter"] = kwargs.get("start_after")

        return params


class DeleteJobArtifacts(GetBaseJobApiUrlMixin, DeleteResource):
    VALIDATION_ERROR_MESSAGE = "Failed to delete resource"

    def get_request_url(self, **kwargs):
        return "/jobs/{}/artifactsDestroy".format(kwargs.get("id"))

    def _send(self, url, **kwargs):
        client = self._get_client(**kwargs)
        params_data = self._get_request_params(kwargs)
        response = self._send_request(client, url, params_data=params_data)
        gradient_response = http_client.GradientResponse.interpret_response(response)
        return gradient_response

    def _send_request(self, client, url, params_data=None):
        response = client.post(url, params=params_data)
        return response

    def _get_request_params(self, kwargs):
        filters = dict()

        if kwargs.get("files"):
            filters["files"] = kwargs.get("files")

        return filters or None


class GetJobArtifacts(GetBaseJobApiUrlMixin, GetResource):
    def _parse_object(self, data, **kwargs):
        return data

    def _get_request_params(self, kwargs):
        return {
            "jobId": kwargs.get("jobId")
        }

    def get_request_url(self, **kwargs):
        return "/jobs/artifactsGet"


class GetJobMetrics(GetMetrics):
    OBJECT_TYPE = "mljob"

    def _get_instance_by_id(self, instance_id, **kwargs):
        repository = GetJob(self.api_key, logger=self.logger, ps_client_name=self.ps_client_name)
        instance = repository.get(job_id=instance_id)
        return instance

    def _get_start_date(self, instance, kwargs):
        rv = super(GetJobMetrics, self)._get_start_date(instance, kwargs)
        if rv is None:
            raise sdk_exceptions.GradientSdkError("Job has not started yet")

        return rv

class ListJobMetrics(ListMetrics):
    OBJECT_TYPE = "mljob"

    def _get_instance_by_id(self, instance_id, **kwargs):
        repository = GetJob(self.api_key, logger=self.logger, ps_client_name=self.ps_client_name)
        instance = repository.get(job_id=instance_id)
        return instance

    def _get_start_date(self, instance, kwargs):
        rv = super(ListJobMetrics, self)._get_start_date(instance, kwargs)
        if rv is None:
            raise sdk_exceptions.GradientSdkError("Job has not started yet")

        return rv

class StreamJobMetrics(StreamMetrics):
    OBJECT_TYPE = "mljob"

    def _get_metrics_api_url(self, instance_id, protocol="https"):
        repository = GetJob(api_key=self.api_key, logger=self.logger, ps_client_name=self.ps_client_name)
        instance = repository.get(job_id=instance_id)

        metrics_api_url = super(StreamJobMetrics, self)._get_metrics_api_url(instance, protocol="wss")
        return metrics_api_url



import marshmallow as ma

from .. import models
from ..serializers.base import BaseSchema


class DatasetVersionSummarySchema(BaseSchema):
    MODEL = models.DatasetVersionTagSummary

    name = ma.fields.Str()


class DatasetVersionSchema(BaseSchema):
    MODEL = models.DatasetVersion

    version = ma.fields.Str()
    message = ma.fields.Str()
    is_committed = ma.fields.Bool(load_from="isCommitted", dump_to="isCommitted")
    tags = ma.fields.Nested(DatasetVersionSummarySchema, load_only=True, many=True)

    dataset_id = ma.fields.Str(dump_to="id", dump_only=True)


class DatasetVersionPreSignedURLSchema(BaseSchema):
    MODEL = models.DatasetVersionPreSignedURL

    url = ma.fields.Str()
    expires_in = ma.fields.Integer()



import curses
import json
import shutil

import requests
import six

from gradient import exceptions


def get_terminal_lines(fallback=48):
    if six.PY3:
        return shutil.get_terminal_size().lines

    return fallback


def print_json_pretty(res):
    print(json.dumps(res, indent=2, sort_keys=True))


def response_error_check(res):
    if ('error' not in res
            and 'status' in res
            and (res['status'] < 200 or res['status'] > 299)):
        res['error'] = True
    return res


def requests_exception_to_error_obj(e):
    return {'error': True, 'message': str(e)}


def status_code_to_error_obj(status_code):
    message = 'unknown'
    if status_code in requests.status_codes._codes:
        message = requests.status_codes._codes[status_code][0]
    return {'error': True, 'message': message, 'status': status_code}


def validate_auth_options(kwargs):
    if kwargs["generate_auth"] and any((kwargs["auth_username"], kwargs["auth_password"])):
        raise exceptions.ApplicationError("Use either --authUsername and --authPassword options or --auth flag")

    # checking if both or none auth options were used
    if len([val for val in (kwargs["auth_username"], kwargs["auth_password"]) if val is not None]) == 1:
        raise exceptions.ApplicationError("--authUsername and --authPassword have to be used together")


def none_strings_to_none_objects(lst):
    return [elem if elem != "none" else None for elem in lst]


class TerminalPrinter(object):
    def __init__(self):
        self.screen_writer = None

    def init(self):
        self.screen_writer = curses.initscr()
        return self

    def cleanup(self):
        """Cleanup before program ends or terminal may stop working correlctly"""
        curses.nocbreak()
        self.screen_writer.keypad(False)
        curses.echo()
        curses.endwin()

    def rewrite_screen(self, s):
        self.clear()
        self.add_line(s)
        self.refresh()

    def add_line(self, s):
        self.add_str(s + "\n")

    def add_str(self, s):
        self.screen_writer.addstr(s)

    def refresh(self):
        self.screen_writer.refresh()

    def clear(self):
        self.screen_writer.clear()



import marshmallow as ma

from .. import models
from ..serializers.base import BaseSchema


class StorageProviderSchema(BaseSchema):
    MODEL = models.StorageProvider

    id = ma.fields.Str()
    type = ma.fields.Str()
    name = ma.fields.Str()
    config = ma.fields.Dict()



import abc
import itertools
import json
import pydoc

import six
import terminaltables
from click import style
from halo import halo

from gradient import exceptions, DeploymentsClient, AutoscalingMetric, AutoscalingDefinition
from gradient.api_sdk import sdk_exceptions, utils
from gradient.api_sdk.config import config
from gradient.api_sdk.utils import concatenate_urls
from gradient.cli_constants import CLI_PS_CLIENT_NAME
from gradient.cliutils import get_terminal_lines
from gradient.commands.common import DetailsCommandMixin, StreamMetricsCommand, BaseCommand, LogsCommandMixin


@six.add_metaclass(abc.ABCMeta)
class BaseDeploymentCommand(BaseCommand):
    def _get_client(self, api_key, logger):
        client = DeploymentsClient(
            api_key=api_key,
            logger=logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )
        return client


class HandleAutoscalingOptions(object):
    def _handle_autoscaling_options(self, kwargs):
        autoscaling_metrics_and_resources = []
        metrics = kwargs.pop("metrics", None) or []
        resources = kwargs.pop("resources", None) or []
        for metric_dict in itertools.chain(resources, metrics):
            metric = AutoscalingMetric(
                type=metric_dict["type"],
                name=metric_dict["name"],
                value_type=metric_dict["value_type"],
                value=metric_dict["value"],
            )
            autoscaling_metrics_and_resources.append(metric)

        autoscaling_definition = AutoscalingDefinition(
            min_instance_count=kwargs.pop("min_instance_count", None),
            max_instance_count=kwargs.pop("max_instance_count", None),
            scale_cooldown_period=kwargs.pop("scale_cooldown_period", None),
            metrics=autoscaling_metrics_and_resources,
        )
        kwargs["autoscaling"] = autoscaling_definition


class HandleWorkspaceMixin(object):
    def _handle_workspace(self, instance_dict):
        handler = self.workspace_handler.handle(instance_dict)

        instance_dict.pop("ignore_files", None)
        instance_dict.pop("workspace", None)
        if handler:
            instance_dict["workspace_url"] = handler


class CreateDeploymentCommand(HandleAutoscalingOptions, BaseDeploymentCommand, HandleWorkspaceMixin):
    def __init__(self, workspace_handler, *args, **kwargs):
        super(CreateDeploymentCommand, self).__init__(*args, **kwargs)
        self.workspace_handler = workspace_handler

    def execute(self, **kwargs):
        self._handle_auth(kwargs)
        self._handle_workspace(kwargs)
        self._handle_autoscaling_options(kwargs)
        with halo.Halo(text="Creating new deployment", spinner="dots"):
            deployment_id = self.client.create(**kwargs)

        self.logger.log("New deployment created with id: {}".format(deployment_id))
        self.logger.log(self.get_instance_url(deployment_id))

    def get_instance_url(self, instance_id):
        url = concatenate_urls(config.WEB_URL, "{}/deployments/{}".format(self.get_namespace(), instance_id))
        return url

    def _handle_auth(self, kwargs):
        if kwargs.pop("generate_auth", False):
            auth_username, auth_password = utils.generate_credentials_pair(12)
            kwargs["auth_username"] = auth_username
            kwargs["auth_password"] = auth_password
            self.logger.log("Generated credentials: \nusername:{}\npassword:{}".format(auth_password, auth_username))


class ListDeploymentsCommand(BaseDeploymentCommand):
    WAITING_FOR_RESPONSE_MESSAGE = "Waiting for data..."

    def execute(self, **kwargs):
        with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
            instances = self._get_instances(**kwargs)

        self._log_objects_list(instances)

    def _get_instances(self, **kwargs):
        try:
            instances = self.client.list(**kwargs)
        except sdk_exceptions.GradientSdkError as e:
            raise exceptions.ReceivingDataFailedError(e)

        return instances

    @staticmethod
    def _get_table_data(deployments):
        data = [("Name", "ID", "Endpoint", "Api Type", "Deployment Type", "Deployment State")]
        for deployment in deployments:
            name = deployment.name
            id_ = deployment.id
            endpoint = deployment.endpoint
            api_type = deployment.api_type
            deployment_type = deployment.deployment_type
            deployment_state = deployment.state
            data.append((name, id_, endpoint, api_type, deployment_type, deployment_state))

        return data

    def _log_objects_list(self, objects):
        if not objects:
            self.logger.warning("No data found")
            return

        table_data = self._get_table_data(objects)
        table_str = self._make_table(table_data)
        if len(table_str.splitlines()) > get_terminal_lines():
            pydoc.pager(table_str)
        else:
            self.logger.log(table_str)

    @staticmethod
    def _make_table(table_data):
        ascii_table = terminaltables.AsciiTable(table_data)
        table_string = ascii_table.table
        return table_string


class StartDeploymentCommand(BaseDeploymentCommand):
    def execute(self, **kwargs):
        self.client.start(**kwargs)
        self.logger.log("Deployment started")


class StopDeploymentCommand(BaseDeploymentCommand):
    def execute(self, **kwargs):
        self.client.stop(**kwargs)
        self.logger.log("Deployment stopped")


class DeleteDeploymentCommand(BaseDeploymentCommand):
    def execute(self, **kwargs):
        self.client.delete(**kwargs)
        self.logger.log("Deployment deleted")


class UpdateDeploymentCommand(HandleAutoscalingOptions, BaseDeploymentCommand, HandleWorkspaceMixin):
    def __init__(self, workspace_handler, *args, **kwargs):
        super(UpdateDeploymentCommand, self).__init__(*args, **kwargs)
        self.workspace_handler = workspace_handler

    def execute(self, deployment_id, **kwargs):
        self._handle_workspace(kwargs)
        self._handle_autoscaling_options(kwargs)

        with halo.Halo(text="Updating deployment data", spinner="dots"):
            self.client.update(deployment_id, **kwargs)

        self.logger.log("Deployment data updated")


class GetDeploymentDetails(DetailsCommandMixin, BaseDeploymentCommand):
    def _get_table_data(self, instance):
        """
        :param models.Deployment instance:
        """
        tags_string = ", ".join(instance.tags)
        autoscaling_metrics_string = self.get_autoscaling_metrics_string(instance)

        data = (
            ("ID", instance.id),
            ("Name", instance.name),
            ("State", instance.state),
            ("Machine type", instance.machine_type),
            ("Instance count", instance.instance_count),
            ("Command", instance.command),
            ("Deployment type", instance.deployment_type),
            ("Model ID", instance.model_id),
            ("Project ID", instance.project_id),
            ("Endpoint", instance.endpoint),
            ("API type", instance.api_type),
            ("Cluster ID", instance.cluster_id),
            ("Tags", tags_string),
            ("Min Instance Count", getattr(instance.autoscaling, "min_instance_count", "")),
            ("Max Instance Count", getattr(instance.autoscaling, "max_instance_count", "")),
            ("Scale Cooldown Period", getattr(instance.autoscaling, "scale_cooldown_period", "")),
            ("Autoscaling Metrics", autoscaling_metrics_string),
        )
        return data

    def get_autoscaling_metrics_string(self, instance):
        if not instance.autoscaling or not instance.autoscaling.metrics:
            return ""

        s = "\n".join("{}/{}:{}".format(m.name, m.value_type, m.value)
                      for m in instance.autoscaling.metrics)
        return s


class DeploymentAddTagsCommand(BaseDeploymentCommand):
    def execute(self, deployment_id, *args, **kwargs):
        self.client.add_tags(deployment_id, **kwargs)
        self.logger.log("Tags added to deployment")


class DeploymentRemoveTagsCommand(BaseDeploymentCommand):
    def execute(self, deployment_id, *args, **kwargs):
        self.client.remove_tags(deployment_id, **kwargs)
        self.logger.log("Tags removed from deployment")


class GetDeploymentMetricsCommand(BaseDeploymentCommand):
    def execute(self, deployment_id, start, end, interval, built_in_metrics, *args, **kwargs):
        metrics = self.client.get_metrics(
            deployment_id,
            start=start,
            end=end,
            built_in_metrics=built_in_metrics,
            interval=interval,
        )
        formatted_metrics = json.dumps(metrics, indent=2, sort_keys=True)
        self.logger.log(formatted_metrics)


class ListDeploymentMetricsCommand(BaseDeploymentCommand):
    def execute(self, deployment_id, start, end, interval, *args, **kwargs):
        metrics = self.client.list_metrics(
            deployment_id,
            start=start,
            end=end,
            interval=interval,
        )
        formatted_metrics = json.dumps(metrics, indent=2, sort_keys=True)
        self.logger.log(formatted_metrics)

class StreamDeploymentMetricsCommand(StreamMetricsCommand, BaseDeploymentCommand):
    pass


class DeploymentLogsCommand(LogsCommandMixin, BaseDeploymentCommand):
    def _make_table(self, logs, experiment_id):
        table_title = "Deployment %s logs" % experiment_id
        table_data = [("LINE", "MESSAGE")]
        table = terminaltables.AsciiTable(table_data, title=table_title)

        for log in logs:
            table_data.append(self._format_row(experiment_id, log))

        return table.table

    def _get_log_row_string(self, id, log):
        log_msg = "{}\t{}".format(*self._format_row(id, log))
        return log_msg

    @staticmethod
    def _format_row(id, log_row):
        return (style(fg="red", text=str(log_row.line)),
                log_row.message)



import abc
import json
import pydoc

import halo
import six
import terminaltables
from click import style

from gradient import api_sdk, exceptions
from gradient.api_sdk import sdk_exceptions
from gradient.cli_constants import CLI_PS_CLIENT_NAME
from gradient.cliutils import get_terminal_lines
from gradient.commands.common import BaseCommand, ListCommandMixin, DetailsCommandMixin, StreamMetricsCommand, \
    LogsCommandMixin


@six.add_metaclass(abc.ABCMeta)
class BaseNotebookCommand(BaseCommand):
    def _get_client(self, api_key, logger):
        client = api_sdk.clients.NotebooksClient(
            api_key=api_key,
            logger=logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )
        return client


class CreateNotebookCommand(BaseNotebookCommand):
    SPINNER_MESSAGE = "Creating new notebook"

    def execute(self, **kwargs):
        with halo.Halo(text=self.SPINNER_MESSAGE, spinner="dots"):
            notebook_id = self.client.create(**kwargs)

        self.logger.log("Created new notebook with id: {}".format(notebook_id))
        self.logger.log(self.get_instance_url(notebook_id))

    def get_instance_url(self, notebook_id):
        notebook = self.client.get(notebook_id)
        return notebook.url


class StopNotebookCommand(BaseNotebookCommand):
    SPINNER_MESSAGE = "Stopping notebook"

    def execute(self, id):
        with halo.Halo(text=self.SPINNER_MESSAGE, spinner="dots"):
            self.client.stop(id)

        self.logger.log("Stopping notebook with id: {}".format(id))


class StartNotebookCommand(BaseNotebookCommand):
    SPINNER_MESSAGE = "Starting notebook"

    def execute(self, **kwargs):
        with halo.Halo(text=self.SPINNER_MESSAGE, spinner="dots"):
            notebook_id = self.client.start(**kwargs)

        self.logger.log("Started notebook with id: {}".format(notebook_id))
        self.logger.log(self.get_instance_url(notebook_id))

    def get_instance_url(self, notebook_id):
        notebook = self.client.get(notebook_id)
        return notebook.url


class ForkNotebookCommand(BaseNotebookCommand):
    WAITING_FOR_RESPONSE_MESSAGE = "Forking notebook"

    def execute(self, id_, project_id):
        with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
            handle = self.client.fork(id_, project_id)

        self.logger.log("Notebook forked to id: {}".format(handle))


class DeleteNotebookCommand(BaseNotebookCommand):
    WAITING_FOR_RESPONSE_MESSAGE = "Deleting notebook"

    def execute(self, id_):
        with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
            self.client.delete(id_)

        self.logger.log("Notebook deleted")


class ListNotebooksCommand(ListCommandMixin, BaseNotebookCommand):
    SPINNER_MESSAGE = "Waiting for data"

    def _get_instances(self, **kwargs):
        limit = kwargs.get("limit")
        offset = kwargs.get("offset")
        tags = kwargs.get("tags")
        get_meta = True
        try:
            instances, meta_data = self.client.list(get_meta=get_meta, limit=limit, offset=offset, tags=tags)
        except sdk_exceptions.GradientSdkError as e:
            raise exceptions.ReceivingDataFailedError(e)
        return instances, meta_data

    def _get_table_data(self, notebooks):
        data = [("Name", "ID")]
        for obj in notebooks:
            data.append((obj.name, obj.id))
        return data

    def execute(self, **kwargs):
        return self._generate_data_table(**kwargs)


class ShowNotebookDetailsCommand(DetailsCommandMixin, BaseNotebookCommand):
    def _get_table_data(self, instance):
        """
        :param api_sdk.Notebook instance:
        """
        tags_string = ", ".join(instance.tags)

        data = (
            ("Name", instance.name),
            ("ID", instance.id),
            ("Project", instance.project_handle),
            ("VM Type", instance.vm_type),
            ("State", instance.state),
            ("FQDN", instance.fqdn),
            ("Tags", tags_string),
        )
        return data


class NotebookAddTagsCommand(BaseNotebookCommand):
    def execute(self, notebook_id, *args, **kwargs):
        self.client.add_tags(notebook_id, **kwargs)
        self.logger.log("Tags added to notebook")


class NotebookRemoveTagsCommand(BaseNotebookCommand):
    def execute(self, notebook_id, *args, **kwargs):
        self.client.remove_tags(notebook_id, **kwargs)
        self.logger.log("Tags removed from notebook")


class GetNotebookMetricsCommand(BaseNotebookCommand):
    def execute(self, notebook_id, start, end, interval, built_in_metrics, *args, **kwargs):
        metrics = self.client.get_metrics(
            notebook_id,
            start=start,
            end=end,
            built_in_metrics=built_in_metrics,
            interval=interval,
        )
        formatted_metrics = json.dumps(metrics, indent=2, sort_keys=True)
        self.logger.log(formatted_metrics)


class ListNotebookMetricsCommand(BaseNotebookCommand):
    def execute(self, notebook_id, start, end, interval, *args, **kwargs):
        metrics = self.client.list_metrics(
            notebook_id,
            start=start,
            end=end,
            interval=interval,
        )
        formatted_metrics = json.dumps(metrics, indent=2, sort_keys=True)
        self.logger.log(formatted_metrics)


class StreamNotebookMetricsCommand(StreamMetricsCommand, BaseNotebookCommand):
    pass


class ArtifactsListCommand(BaseNotebookCommand):
    WAITING_FOR_RESPONSE_MESSAGE = "Waiting for data..."

    def execute(self, **kwargs):
        with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
            try:
                instances = self.client.artifacts_list(**kwargs)
            except sdk_exceptions.GradientSdkError as e:
                raise exceptions.ReceivingDataFailedError(e)

        self._log_objects_list(instances, kwargs)

    def _get_table_data(self, artifacts, kwargs):
        columns = ['Files']

        show_size = "size" in kwargs
        show_url = "url" in kwargs

        if show_size:
            columns.append('Size (in bytes)')
        if show_url:
            columns.append('URL')

        data = [tuple(columns)]
        for artifact in artifacts:
            row = [artifact.file]
            if show_size:
                row.append(artifact.size)
            if show_url:
                row.append(artifact.url)
            data.append(tuple(row))
        return data

    def _log_objects_list(self, objects, kwargs):
        if not objects:
            self.logger.warning("No data found")
            return

        table_data = self._get_table_data(objects, kwargs)
        table_str = self._make_table(table_data)
        if len(table_str.splitlines()) > get_terminal_lines():
            pydoc.pager(table_str)
        else:
            self.logger.log(table_str)

    @staticmethod
    def _make_table(table_data):
        ascii_table = terminaltables.AsciiTable(table_data)
        table_string = ascii_table.table
        return table_string


class NotebookLogsCommand(LogsCommandMixin, BaseNotebookCommand):
    ENTITY = "Notebook"



import abc

import six
import websocket

from .common import ListResources, CreateResource, StartResource, StopResource, DeleteResource, GetResource, GetMetrics, ListMetrics, \
    StreamMetrics, ListLogs
from .. import config, serializers, sdk_exceptions
from ..repositories.jobs import ListJobs
from ..serializers import utils as serializers_utils
from ..utils import concatenate_urls


class GetBaseExperimentApiUrlMixin(object):
    def _get_api_url(self, **_):
        return config.config.CONFIG_EXPERIMENTS_HOST_V2


class GetBaseExperimentApiUrlBasedOnClusterIdMixin(object):
    def _get_api_url(self, **kwargs):
        if kwargs.get("clusterId"):
            return config.config.CONFIG_EXPERIMENTS_HOST_V2

        return config.config.CONFIG_EXPERIMENTS_HOST


class ParseExperimentDictMixin(object):
    def _parse_object(self, experiment_dict, **kwargs):
        """
        :param dict experiment_dict:
        :rtype BaseExperiment
        """
        serializer = serializers_utils.get_serializer_for_experiment(experiment_dict)
        experiment = serializer().get_instance(experiment_dict)
        return experiment


class ListExperiments(ParseExperimentDictMixin, GetBaseExperimentApiUrlMixin, ListResources):
    def get_request_url(self, **kwargs):
        return "/experiments/"

    def _get_meta_data(self, resp):
        return resp.data.get("meta")

    def _parse_objects(self, data, **kwargs):
        experiments_dicts = self._get_experiments_dicts_from_json_data(data, kwargs)
        experiments = []
        for experiment_dict in experiments_dicts:
            experiment_dict.update(experiment_dict["templateHistory"].get("params", {}))
            experiment = self._parse_object(experiment_dict)
            experiments.append(experiment)

        return experiments

    @staticmethod
    def _get_experiments_dicts_from_json_data(data, kwargs):
        filtered = bool(kwargs.get("project_id"))
        if not filtered:  # If filtering by project ID response data has different format...
            return data["data"]

        experiments = []
        for project_experiments in data["data"]:
            for experiment in project_experiments["data"]:
                experiments.append(experiment)

        return experiments

    def _get_request_params(self, kwargs):
        params = {
            "limit": kwargs.get("limit"),
            "offset": kwargs.get("offset")
        }

        project_id = kwargs.get("project_id")
        if project_id:
            if isinstance(project_id, six.string_types):
                project_id = [project_id]
            for i, experiment_id in enumerate(project_id):
                key = "projectHandle[{}]".format(i)
                params[key] = experiment_id

        tags = kwargs.get("tags")
        if tags:
            params["tag"] = tags

        return params


class GetExperiment(ParseExperimentDictMixin, GetBaseExperimentApiUrlMixin, GetResource):
    def _parse_object(self, experiment_dict, **kwargs):
        experiment_dict = experiment_dict["data"]
        experiment_dict.update(experiment_dict["templateHistory"].get("params", {}))
        return super(GetExperiment, self)._parse_object(experiment_dict, **kwargs)

    def get_request_url(self, **kwargs):
        experiment_id = kwargs["experiment_id"]
        url = "/experiments/{}/".format(experiment_id)
        return url


class ListExperimentLogs(ListLogs):
    def _get_request_params(self, kwargs):
        params = {
            "experimentId": kwargs["id"],
            "line": kwargs["line"],
            "limit": kwargs["limit"],
        }
        return params


@six.add_metaclass(abc.ABCMeta)
class BaseCreateExperiment(GetBaseExperimentApiUrlBasedOnClusterIdMixin, CreateResource):
    def get_request_url(self, **_):
        return "/experiments/"


class CreateSingleNodeExperiment(BaseCreateExperiment):
    SERIALIZER_CLS = serializers.SingleNodeExperimentSchema


class CreateMultiNodeExperiment(BaseCreateExperiment):
    SERIALIZER_CLS = serializers.MultiNodeExperimentSchema


class CreateMpiMultiNodeExperiment(BaseCreateExperiment):
    SERIALIZER_CLS = serializers.MpiMultiNodeExperimentSchema


class RunSingleNodeExperiment(CreateSingleNodeExperiment):
    def get_request_url(self, **_):
        return "/experiments/run/"


class RunMultiNodeExperiment(CreateMultiNodeExperiment):
    def get_request_url(self, **_):
        return "/experiments/run/"


class RunMpiMultiNodeExperiment(CreateMpiMultiNodeExperiment):
    def get_request_url(self, **_):
        return "/experiments/run/"


class StartExperiment(GetBaseExperimentApiUrlMixin, StartResource):
    VALIDATION_ERROR_MESSAGE = "Failed to start experiment"

    def get_request_url(self, **kwargs):
        id_ = kwargs["id"]
        url = "/experiments/{}/start/".format(id_)
        return url


class StopExperiment(GetBaseExperimentApiUrlMixin, StopResource):
    VALIDATION_ERROR_MESSAGE = "Failed to stop experiment"

    def get_request_url(self, **kwargs):
        id_ = kwargs["id"]
        url = "/experiments/{}/stop/".format(id_)
        return url


class DeleteExperiment(GetBaseExperimentApiUrlMixin, DeleteResource):
    def get_request_url(self, **kwargs):
        experiment_id = kwargs["id"]
        return "/experiments/{}/".format(experiment_id)


class GetExperimentMetricsApiUrlMixin(object):
    def _get_metrics_api_url(self, instance, protocol="https"):
        instance_id = getattr(instance, "id", instance)
        repository = ListJobs(api_key=self.api_key, logger=self.logger, ps_client_name=self.ps_client_name)
        try:
            job = repository.list(experiment_id=instance_id)[0]
        except IndexError:
            raise sdk_exceptions.GradientSdkError("Experiment has not started yet")

        metrics_api_url = concatenate_urls(protocol + "://", job.metrics_url)
        if not job.metrics_url:
            raise sdk_exceptions.GradientSdkError("Metrics API url not found")

        return metrics_api_url


class GetExperimentMetrics(GetExperimentMetricsApiUrlMixin, GetMetrics):
    OBJECT_TYPE = "experiment"

    def _get_instance_by_id(self, instance_id, **kwargs):
        repository = GetExperiment(self.api_key, logger=self.logger, ps_client_name=self.ps_client_name)
        instance = repository.get(experiment_id=instance_id)
        return instance

    def _get_start_date(self, instance, kwargs):
        rv = super(GetExperimentMetrics, self)._get_start_date(instance, kwargs)
        if rv is None:
            raise sdk_exceptions.GradientSdkError("Experiment has not started yet")

        return rv

    def _get_instance(self, response, **kwargs):
        try:
            rv = super(GetExperimentMetrics, self)._get_instance(response, **kwargs)
        except sdk_exceptions.ResourceFetchingError as e:
            if '{"version":' in str(e):
                # TODO: metrics are not working for v1 experiments at the moment
                raise sdk_exceptions.GradientSdkError("Metrics are available for private clusters only")
            else:
                raise

        return rv

class ListExperimentMetrics(GetExperimentMetricsApiUrlMixin, ListMetrics):
    OBJECT_TYPE = "experiment"

    def _get_instance_by_id(self, instance_id, **kwargs):
        repository = GetExperiment(self.api_key, logger=self.logger, ps_client_name=self.ps_client_name)
        instance = repository.get(experiment_id=instance_id)
        return instance

    def _get_start_date(self, instance, kwargs):
        rv = super(ListExperimentMetrics, self)._get_start_date(instance, kwargs)
        if rv is None:
            raise sdk_exceptions.GradientSdkError("Experiment has not started yet")

        return rv

    def _get_instance(self, response, **kwargs):
        try:
            rv = super(ListExperimentMetrics, self)._get_instance(response, **kwargs)
        except sdk_exceptions.ResourceFetchingError as e:
            if '{"version":' in str(e):
                # TODO: metrics are not working for v1 experiments at the moment
                raise sdk_exceptions.GradientSdkError("Custom metrics are available for private clusters only")
            else:
                raise

        return rv

class StreamExperimentMetrics(GetExperimentMetricsApiUrlMixin, StreamMetrics):
    OBJECT_TYPE = "experiment"

    def _get_connection(self, kwargs):
        try:
            ws = super(StreamExperimentMetrics, self)._get_connection(kwargs)
        except websocket.WebSocketBadStatusException as e:
            if "Handshake status 200 OK" in str(e):
                # TODO: metrics are not working for v1 experiments at the moment
                raise sdk_exceptions.GradientSdkError("Metrics are available for private clusters only")
            else:
                raise

        return ws
